# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from astrapy.utils.api_options import FullTimeoutOptions


def _first_valid_timeout(
    *items: tuple[int | None, str | None],
) -> tuple[int, str | None]:
    # items are: (timeout ms, label)
    not_nulls = [itm for itm in items if itm[0] is not None]
    if not_nulls:
        return not_nulls[0]  # type: ignore[return-value]
    else:
        # If no non-nulls provided, return zero
        # (a timeout of zero will stand for 'no timeout' later on to the request).
        return 0, None


def _min_labeled_timeout(
    *timeouts: tuple[int | None, str | None],
) -> tuple[int, str | None]:
    _non_null: list[tuple[int, str | None]] = [
        to  # type: ignore[misc]
        for to in timeouts
        if to[0] is not None
    ]
    if _non_null:
        min_to, min_lb = min(_non_null, key=lambda p: p[0])
        # min_to is never None, this is for added robustness
        return (min_to or 0, min_lb)
    else:
        return (0, None)


def _select_singlereq_timeout_gm(
    *,
    timeout_options: FullTimeoutOptions,
    general_method_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (non-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (general_method_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.general_method_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "general_method_timeout_ms")
    else:
        return _min_labeled_timeout(
            (general_method_timeout_ms, "general_method_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )


def _select_singlereq_timeout_ca(
    *,
    timeout_options: FullTimeoutOptions,
    collection_admin_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (collection-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (collection_admin_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.collection_admin_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "collection_admin_timeout_ms")
    else:
        return _min_labeled_timeout(
            (collection_admin_timeout_ms, "collection_admin_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )


def _select_singlereq_timeout_ta(
    *,
    timeout_options: FullTimeoutOptions,
    table_admin_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (table-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (table_admin_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.table_admin_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "table_admin_timeout_ms")
    else:
        return _min_labeled_timeout(
            (table_admin_timeout_ms, "table_admin_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )


def _select_singlereq_timeout_da(
    *,
    timeout_options: FullTimeoutOptions,
    database_admin_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (database-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (database_admin_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.database_admin_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "database_admin_timeout_ms")
    else:
        return _min_labeled_timeout(
            (database_admin_timeout_ms, "database_admin_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )


def _select_singlereq_timeout_ka(
    *,
    timeout_options: FullTimeoutOptions,
    keyspace_admin_timeout_ms: int | None,
    request_timeout_ms: int | None = None,
    timeout_ms: int | None = None,
) -> tuple[int, str | None]:
    """
    Apply the logic for determining and labeling the timeout for
    (keyspace-admin) single-request methods.

    If no int args are passed, pick (and correctly label) the least of the
    involved parameters.
    If any of the int args are passed, pick (and correctly labeled) the least
    of them, disregarding the options altogether.
    """
    if all(
        iarg is None
        for iarg in (keyspace_admin_timeout_ms, request_timeout_ms, timeout_ms)
    ):
        ao_r = timeout_options.request_timeout_ms
        ao_gm = timeout_options.keyspace_admin_timeout_ms
        if ao_r < ao_gm:
            return (ao_r, "request_timeout_ms")
        else:
            return (ao_gm, "keyspace_admin_timeout_ms")
    else:
        return _min_labeled_timeout(
            (keyspace_admin_timeout_ms, "keyspace_admin_timeout_ms"),
            (request_timeout_ms, "request_timeout_ms"),
            (timeout_ms, "timeout_ms"),
        )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import httpx

from astrapy.exceptions.error_descriptors import (
    DataAPIErrorDescriptor,
    DataAPIWarningDescriptor,
)


class DataAPIException(Exception):
    """
    Any exception occurred while issuing requests to the Data API
    and specific to it, such as:
      - a collection is found not to exist when gettings its metadata,
      - the API return a response with an error,
    but not, for instance,
      - a network error while sending an HTTP request to the API.
    """

    pass


@dataclass
class DataAPIResponseException(DataAPIException):
    """
    The Data API returned an HTTP 200 ("success") response, which however
    reports API-specific error(s), possibly alongside partial successes.

    Attributes:
        text: a text message about the exception.
        command: the payload to the API that led to the response.
        raw_response: the full response from the API.
        error_descriptors: a list of DataAPIErrorDescriptor, one for each
            item in the API response's "errors" field.
        warning_descriptors: a list of DataAPIWarningDescriptor, one for each
            item in the API response's "warnings" field (if there are any).
    """

    text: str | None
    command: dict[str, Any] | None
    raw_response: dict[str, Any]
    error_descriptors: list[DataAPIErrorDescriptor]
    warning_descriptors: list[DataAPIWarningDescriptor]

    def __init__(
        self,
        text: str | None,
        *,
        command: dict[str, Any] | None,
        raw_response: dict[str, Any],
        error_descriptors: list[DataAPIErrorDescriptor],
        warning_descriptors: list[DataAPIWarningDescriptor],
    ) -> None:
        super().__init__(text)
        self.text = text
        self.command = command
        self.raw_response = raw_response
        self.error_descriptors = error_descriptors
        self.warning_descriptors = warning_descriptors

    @staticmethod
    def from_response(
        *,
        command: dict[str, Any] | None,
        raw_response: dict[str, Any],
        **kwargs: Any,
    ) -> DataAPIResponseException:
        """Parse a raw response from the API into this exception."""

        error_descriptors = [
            DataAPIErrorDescriptor(error_dict)
            for error_dict in (raw_response or {}).get("errors") or []
        ]
        warning_descriptors = [
            DataAPIWarningDescriptor(error_dict)
            for error_dict in (raw_response or {}).get("warnings") or []
        ]

        if error_descriptors:
            summaries = [e_d.summary() for e_d in error_descriptors]
            if len(summaries) == 1:
                text = summaries[0]
            else:
                _j_summaries = "; ".join(
                    f"[{summ_i + 1}] {summ_s}"
                    for summ_i, summ_s in enumerate(summaries)
                )
                text = f"[{len(summaries)} errors collected] {_j_summaries}"
        else:
            text = ""

        return DataAPIResponseException(
            text,
            command=command,
            raw_response=raw_response,
            error_descriptors=error_descriptors,
            warning_descriptors=warning_descriptors,
            **kwargs,
        )


@dataclass
class DataAPIHttpException(DataAPIException, httpx.HTTPStatusError):
    """
    A request to the Data API resulted in an HTTP 4xx or 5xx response.

    In most cases this comes with additional information: the purpose
    of this class is to present such information in a structured way,
    akin to what happens for the DataAPIResponseException, while
    still raising (a subclass of) `httpx.HTTPStatusError`.

    Attributes:
        text: a text message about the exception.
        error_descriptors: a list of all DataAPIErrorDescriptor objects
            found in the response.
    """

    text: str | None
    error_descriptors: list[DataAPIErrorDescriptor]

    def __init__(
        self,
        text: str | None,
        *,
        httpx_error: httpx.HTTPStatusError,
        error_descriptors: list[DataAPIErrorDescriptor],
    ) -> None:
        DataAPIException.__init__(self, text)
        httpx.HTTPStatusError.__init__(
            self,
            message=str(httpx_error),
            request=httpx_error.request,
            response=httpx_error.response,
        )
        self.text = text
        self.httpx_error = httpx_error
        self.error_descriptors = error_descriptors

    def __str__(self) -> str:
        return self.text or str(self.httpx_error)

    @classmethod
    def from_httpx_error(
        cls,
        httpx_error: httpx.HTTPStatusError,
        **kwargs: Any,
    ) -> DataAPIHttpException:
        """Parse a httpx status error into this exception."""

        raw_response: dict[str, Any]
        # the attempt to extract a response structure cannot afford failure.
        try:
            raw_response = httpx_error.response.json() or {}
        except Exception:
            raw_response = {}
        error_descriptors = [
            DataAPIErrorDescriptor(error_dict)
            for error_dict in raw_response.get("errors") or []
        ]
        if error_descriptors:
            text = f"{error_descriptors[0].message}. {str(httpx_error)}"
        else:
            text = str(httpx_error)

        return cls(
            text=text,
            httpx_error=httpx_error,
            error_descriptors=error_descriptors,
            **kwargs,
        )


@dataclass
class DataAPITimeoutException(DataAPIException):
    """
    A Data API operation timed out. This can be a request timeout occurring
    during a specific HTTP request, or can happen over the course of a method
    involving several requests in a row, such as a paginated find.

    Attributes:
        text: a textual description of the error
        timeout_type: this denotes the phase of the HTTP request when the event
            occurred ("connect", "read", "write", "pool") or "generic" if there is
            not a specific request associated to the exception.
        endpoint: if the timeout is tied to a specific request, this is the
            URL that the request was targeting.
        raw_payload:  if the timeout is tied to a specific request, this is the
            associated payload (as a string).
    """

    text: str
    timeout_type: str
    endpoint: str | None
    raw_payload: str | None

    def __init__(
        self,
        text: str,
        *,
        timeout_type: str,
        endpoint: str | None,
        raw_payload: str | None,
    ) -> None:
        super().__init__(text)
        self.text = text
        self.timeout_type = timeout_type
        self.endpoint = endpoint
        self.raw_payload = raw_payload


@dataclass
class CursorException(DataAPIException):
    """
    The cursor operation cannot be invoked if a cursor is not in its pristine
    state (i.e. is already being consumed or is exhausted altogether).

    Attributes:
        text: a text message about the exception.
        cursor_state: a string description of the current state
            of the cursor. See the documentation for Cursor.
    """

    text: str
    cursor_state: str

    def __init__(
        self,
        text: str,
        *,
        cursor_state: str,
    ) -> None:
        super().__init__(text)
        self.text = text
        self.cursor_state = cursor_state


@dataclass
class UnexpectedDataAPIResponseException(DataAPIException):
    """
    The Data API response is malformed in that it does not have
    expected field(s), or they are of the wrong type.

    Attributes:
        text: a text message about the exception.
        raw_response: the response returned by the API in the form of a dict.
    """

    text: str
    raw_response: dict[str, Any] | None

    def __init__(
        self,
        text: str,
        raw_response: dict[str, Any] | None,
    ) -> None:
        super().__init__(text)
        self.text = text
        self.raw_response = raw_response



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

import pytest

from astrapy.api_options import APIOptions, SerdesOptions
from astrapy.constants import SortMode
from astrapy.cursors import CursorState
from astrapy.data_types import DataAPIVector
from astrapy.exceptions import CursorException

from ..conftest import DefaultAsyncCollection

NUM_DOCS = 25  # keep this between 20 and 39
NUM_DOCS_PAGINATION = 90  # keep this above 2 * (2 * 20) and below 2 * (3 * 20)


@pytest.fixture
async def async_filled_collection(
    async_empty_collection: DefaultAsyncCollection,
) -> DefaultAsyncCollection:
    await async_empty_collection.insert_many(
        [
            {
                "p_text": "pA",
                "p_int": i,
                "$vector": DataAPIVector([i, 1]),
            }
            for i in range(NUM_DOCS)
        ]
    )
    return async_empty_collection


@pytest.fixture
async def async_filled_pagination_collection(
    async_empty_collection: DefaultAsyncCollection,
) -> DefaultAsyncCollection:
    await async_empty_collection.insert_many(
        [
            {
                "_id": i,
                "text": f"doc number {i}",
                "even": i % 2 == 0,
                "$vector": DataAPIVector([i, 1]),
            }
            for i in range(NUM_DOCS_PAGINATION)
        ]
    )
    return async_empty_collection


class TestCollectionCursorSync:
    @pytest.mark.describe("test of an IDLE collection cursors properties, async")
    async def test_collection_cursors_idle_properties_async(
        self,
        async_filled_collection: DefaultAsyncCollection,
    ) -> None:
        cur = async_filled_collection.find()
        assert cur.state == CursorState.IDLE

        assert cur.data_source == async_filled_collection

        assert cur.consumed == 0
        assert cur.consume_buffer(3) == []
        assert cur.buffered_count == 0
        assert cur.consumed == 0

        toclose = cur.clone()
        toclose.close()
        toclose.close()
        assert toclose.state == CursorState.CLOSED
        with pytest.raises(CursorException):
            async for row in toclose:
                pass
        with pytest.raises(StopAsyncIteration):
            await toclose.__anext__()
        with pytest.raises(CursorException):
            await toclose.for_each(lambda row: None)
        with pytest.raises(CursorException):
            await toclose.to_list()

        cur.rewind()
        assert cur.state == CursorState.IDLE
        assert cur.consumed == 0
        assert cur.buffered_count == 0

        cur.filter({"c": True})
        cur.project({"c": True})
        cur.sort({"c": SortMode.ASCENDING})
        cur.limit(1)
        cur.include_similarity(False)
        cur.include_sort_vector(False)
        cur.skip(1)
        cur.map(lambda rw: None)

        cur.project({}).map(lambda rw: None)
        with pytest.raises(CursorException):
            cur.map(lambda rw: None).project({})

    @pytest.mark.describe("test of a CLOSED collection cursors properties, async")
    async def test_collection_cursors_closed_properties_async(
        self,
        async_filled_collection: DefaultAsyncCollection,
    ) -> None:
        cur0 = async_filled_collection.find()
        cur0.close()
        cur0.rewind()
        assert cur0.state == CursorState.IDLE

        cur1 = async_filled_collection.find()
        assert cur1.consumed == 0
        [doc async for doc in cur1]
        assert cur1.state == CursorState.CLOSED
        assert cur1.consume_buffer(2) == []
        assert cur1.consumed == NUM_DOCS
        assert cur1.buffered_count == 0
        cloned = cur1.clone()
        assert cloned.consumed == 0
        assert cloned.buffered_count == 0
        assert cloned.state == CursorState.IDLE

        with pytest.raises(CursorException):
            cur1.filter({"c": True})
        with pytest.raises(CursorException):
            cur1.project({"c": True})
        with pytest.raises(CursorException):
            cur1.sort({"c": SortMode.ASCENDING})
        with pytest.raises(CursorException):
            cur1.limit(1)
        with pytest.raises(CursorException):
            cur1.include_similarity(False)
        with pytest.raises(CursorException):
            cur1.include_sort_vector(False)
        with pytest.raises(CursorException):
            cur1.skip(1)
        with pytest.raises(CursorException):
            cur1.map(lambda rw: None)

    @pytest.mark.describe("test of a STARTED collection cursors properties, async")
    async def test_collection_cursors_started_properties_async(
        self,
        async_filled_collection: DefaultAsyncCollection,
    ) -> None:
        cur = async_filled_collection.find()
        await cur.__anext__()
        # now this has 19 items in buffer, one is consumed
        assert cur.consumed == 1
        assert cur.buffered_count == 19
        assert len(cur.consume_buffer(3)) == 3
        assert cur.consumed == 4
        assert cur.buffered_count == 16
        # from time to time the buffer is empty:
        for _ in range(16):
            await cur.__anext__()
        assert cur.buffered_count == 0
        assert cur.consume_buffer(3) == []
        assert cur.consumed == 20
        assert cur.buffered_count == 0

        with pytest.raises(CursorException):
            cur.filter({"c": True})
        with pytest.raises(CursorException):
            cur.project({"c": True})
        with pytest.raises(CursorException):
            cur.sort({"c": SortMode.ASCENDING})
        with pytest.raises(CursorException):
            cur.limit(1)
        with pytest.raises(CursorException):
            cur.include_similarity(False)
        with pytest.raises(CursorException):
            cur.include_sort_vector(False)
        with pytest.raises(CursorException):
            cur.skip(1)
        with pytest.raises(CursorException):
            cur.map(lambda rw: None)

    @pytest.mark.describe("test of collection cursors has_next, async")
    async def test_collection_cursors_has_next_async(
        self,
        async_filled_collection: DefaultAsyncCollection,
    ) -> None:
        cur = async_filled_collection.find()
        assert cur.state == CursorState.IDLE
        assert cur.consumed == 0
        assert cur.has_next()
        assert cur.state == CursorState.IDLE
        assert cur.consumed == 0
        [doc async for doc in cur]
        assert cur.consumed == NUM_DOCS
        assert cur.state == CursorState.CLOSED  # type: ignore[comparison-overlap]

        curmf = async_filled_collection.find()
        await curmf.__anext__()
        await curmf.__anext__()
        assert curmf.consumed == 2
        assert curmf.state == CursorState.STARTED
        assert curmf.has_next()
        assert curmf.consumed == 2
        assert curmf.state == CursorState.STARTED
        for _ in range(18):
            await curmf.__anext__()
        assert await curmf.has_next()
        assert curmf.consumed == 20
        assert curmf.state == CursorState.STARTED
        assert curmf.buffered_count == NUM_DOCS - 20

        cur0 = async_filled_collection.find()
        cur0.close()
        assert not await cur0.has_next()

    @pytest.mark.describe("test of collection cursors zero matches, async")
    async def test_collection_cursors_zeromatches_async(
        self,
        async_filled_collection: DefaultAsyncCollection,
    ) -> None:
        cur = async_filled_collection.find({"p_text": "ZZ"})
        assert not await cur.has_next()
        assert [doc async for doc in cur] == []

    @pytest.mark.describe("test of prematurely closing collection cursors, async")
    async def test_collection_cursors_early_closing_async(
        self,
        async_filled_collection: DefaultAsyncCollection,
    ) -> None:
        cur = async_filled_collection.find()
        for _ in range(12):
            await cur.__anext__()
        cur.close()
        assert cur.state == CursorState.CLOSED
        assert cur.buffered_count == 0
        assert cur.consumed == 12
        # rewind test
        cur.rewind()
        assert len([doc async for doc in cur]) == NUM_DOCS

    @pytest.mark.describe("test of mappings with collection cursors, async")
    async def test_collection_cursors_mapping_async(
        self,
        async_filled_collection: DefaultAsyncCollection,
    ) -> None:
        base_rows = [doc async for doc in async_filled_collection.find()]
        assert len(base_rows) == NUM_DOCS

        def mint(row: dict[str, Any]) -> int:
            return row["p_int"]  # type: ignore[no-any-return]

        def mmult(val: int) -> int:
            return 1000 * val

        # map, base
        mcur = async_filled_collection.find().map(mint)
        mints = [val async for val in mcur]
        assert mints == [mint(row) for row in base_rows]

        # map composition
        mmcur = async_filled_collection.find().map(mint).map(mmult)
        mmints = [val async for val in mmcur]
        assert mmints == [mmult(mint(row)) for row in base_rows]

        # consume_buffer skips the map
        hmcur = async_filled_collection.find().map(mint)
        for _ in range(10):
            await hmcur.__anext__()
        conbuf = hmcur.consume_buffer(2)
        assert len(conbuf) == 2
        assert all(isinstance(itm, dict) for itm in conbuf)

        # rewind preserves the mapping
        rwcur = async_filled_collection.find().map(mint)
        for _ in range(10):
            await rwcur.__anext__()
        rwcur.rewind()
        assert await rwcur.__anext__() == mints[0]

        # clone rewinds
        cl_unmapped = rwcur.clone()
        assert await cl_unmapped.__anext__() == mint(base_rows[0])

    @pytest.mark.describe("test of collection cursors, for_each and to_list, async")
    async def test_collection_cursors_collective_methods_async(
        self,
        async_filled_collection: DefaultAsyncCollection,
    ) -> None:
        base_rows = [doc async for doc in async_filled_collection.find()]

        # full to_list
        tl_cur = async_filled_collection.find()
        assert await tl_cur.to_list() == base_rows
        assert tl_cur.state == CursorState.CLOSED

        # partially-consumed to_list
        ptl_cur = async_filled_collection.find()
        for _ in range(15):
            await ptl_cur.__anext__()
        assert await ptl_cur.to_list() == base_rows[15:]
        assert ptl_cur.state == CursorState.CLOSED

        # mapped to_list

        def mint(row: dict[str, Any]) -> int:
            return row["p_int"]  # type: ignore[no-any-return]

        mtl_cur = async_filled_collection.find().map(mint)
        for _ in range(13):
            await mtl_cur.__anext__()
        assert await mtl_cur.to_list() == [mint(row) for row in base_rows[13:]]
        assert mtl_cur.state == CursorState.CLOSED

        # full for_each
        accum0: list[dict[str, Any]] = []

        def marker0(row: dict[str, Any], acc: list[dict[str, Any]] = accum0) -> None:
            acc += [row]

        fe_cur = async_filled_collection.find()
        await fe_cur.for_each(marker0)
        assert accum0 == base_rows
        assert fe_cur.state == CursorState.CLOSED

        # full for_each, coroutine
        aaccum0: list[dict[str, Any]] = []

        async def amarker0(
            row: dict[str, Any],
            acc: list[dict[str, Any]] = aaccum0,
        ) -> None:
            acc += [row]

        afe_cur = async_filled_collection.find()
        await afe_cur.for_each(amarker0)
        assert aaccum0 == base_rows
        assert afe_cur.state == CursorState.CLOSED

        # partially-consumed for_each
        accum1: list[dict[str, Any]] = []

        def marker1(row: dict[str, Any], acc: list[dict[str, Any]] = accum1) -> None:
            acc += [row]

        pfe_cur = async_filled_collection.find()
        for _ in range(11):
            await pfe_cur.__anext__()
        await pfe_cur.for_each(marker1)
        assert accum1 == base_rows[11:]
        assert pfe_cur.state == CursorState.CLOSED

        # partially-consumed for_each, coroutine
        aaccum1: list[dict[str, Any]] = []

        async def amarker1(
            row: dict[str, Any],
            acc: list[dict[str, Any]] = aaccum1,
        ) -> None:
            acc += [row]

        apfe_cur = async_filled_collection.find()
        for _ in range(11):
            await apfe_cur.__anext__()
        await apfe_cur.for_each(amarker1)
        assert aaccum1 == base_rows[11:]
        assert apfe_cur.state == CursorState.CLOSED

        # mapped for_each
        accum2: list[int] = []

        def marker2(val: int, acc: list[int] = accum2) -> None:
            acc += [val]

        mfe_cur = async_filled_collection.find().map(mint)
        for _ in range(17):
            await mfe_cur.__anext__()
        await mfe_cur.for_each(marker2)
        assert accum2 == [mint(row) for row in base_rows[17:]]
        assert mfe_cur.state == CursorState.CLOSED

        # mapped for_each, coroutine
        aaccum2: list[int] = []

        async def amarker2(val: int, acc: list[int] = aaccum2) -> None:
            acc += [val]

        amfe_cur = async_filled_collection.find().map(mint)
        for _ in range(17):
            await amfe_cur.__anext__()
        await amfe_cur.for_each(amarker2)
        assert aaccum2 == [mint(row) for row in base_rows[17:]]
        assert amfe_cur.state == CursorState.CLOSED

        # breaking (early) for_each
        accum3: list[dict[str, Any]] = []

        def marker3(row: dict[str, Any], acc: list[dict[str, Any]] = accum3) -> bool:
            acc += [row]
            return len(acc) < 5

        bfe_cur = async_filled_collection.find()
        await bfe_cur.for_each(marker3)
        assert accum3 == base_rows[:5]
        assert bfe_cur.state == CursorState.STARTED
        bfe_another = await bfe_cur.__anext__()
        assert bfe_another == base_rows[5]

        # breaking (early) for_each, coroutine
        aaccum3: list[dict[str, Any]] = []

        async def amarker3(
            row: dict[str, Any],
            acc: list[dict[str, Any]] = aaccum3,
        ) -> bool:
            acc += [row]
            return len(acc) < 5

        abfe_cur = async_filled_collection.find()
        await abfe_cur.for_each(amarker3)
        assert aaccum3 == base_rows[:5]
        assert abfe_cur.state == CursorState.STARTED
        abfe_another = await abfe_cur.__anext__()
        assert abfe_another == base_rows[5]

        # nonbool-nonbreaking for_each
        accum4: list[dict[str, Any]] = []

        def marker4(row: dict[str, Any], acc: list[dict[str, Any]] = accum4) -> int:
            acc += [row]
            return 8 if len(acc) < 5 else 0

        nbfe_cur = async_filled_collection.find()
        await nbfe_cur.for_each(marker4)  # type: ignore[arg-type]
        assert accum4 == base_rows
        assert nbfe_cur.state == CursorState.CLOSED

        # nonbool-nonbreaking for_each, coroutine
        aaccum4: list[dict[str, Any]] = []

        async def amarker4(
            row: dict[str, Any],
            acc: list[dict[str, Any]] = aaccum4,
        ) -> int:
            acc += [row]
            return 8 if len(acc) < 5 else 0

        anbfe_cur = async_filled_collection.find()
        await anbfe_cur.for_each(amarker4)  # type: ignore[arg-type]
        assert aaccum4 == base_rows
        assert anbfe_cur.state == CursorState.CLOSED

    @pytest.mark.describe("test of collection cursors, serdes options obeyance, async")
    async def test_collection_cursors_serdes_options_async(
        self,
        async_filled_collection: DefaultAsyncCollection,
    ) -> None:
        # serdes options obeyance check
        noncustom_compo_collection = async_filled_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False),
            )
        )
        custom_compo_collection = async_filled_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True),
            )
        )
        noncustom_rows = await noncustom_compo_collection.find(
            {},
            projection={"$vector": True},
        ).to_list()
        assert len(noncustom_rows) == NUM_DOCS
        assert all(isinstance(ncrow["$vector"], list) for ncrow in noncustom_rows)
        custom_rows = await custom_compo_collection.find(
            {},
            projection={"$vector": True},
        ).to_list()
        assert len(custom_rows) == NUM_DOCS
        assert all(isinstance(crow["$vector"], DataAPIVector) for crow in custom_rows)

    @pytest.mark.describe("test of collection cursors, initial_page_state, async")
    async def test_collection_cursors_initialpagestate_async(
        self,
        async_filled_pagination_collection: DefaultAsyncCollection,
    ) -> None:
        page_size = 20

        cur0 = async_filled_pagination_collection.find(filter={"even": True})
        ids0: list[int] = []
        for _ in range(page_size):
            doc = await cur0.__anext__()
            ids0.append(doc["_id"])
        nps0 = cur0._next_page_state
        assert isinstance(nps0, str)

        cur1 = async_filled_pagination_collection.find(
            filter={"even": True},
            initial_page_state=nps0,
        )
        ids1: list[int] = []
        for _ in range(page_size):
            doc = await cur1.__anext__()
            ids1.append(doc["_id"])
        nps1 = cur1._next_page_state
        assert isinstance(nps1, str)

        cur2 = async_filled_pagination_collection.find(
            filter={"even": True},
            initial_page_state=nps1,
        )
        ids2 = [doc["_id"] async for doc in cur2]
        assert cur2._next_page_state is None

        expected_ids = [i for i in range(NUM_DOCS_PAGINATION) if i % 2 == 0]
        retrieved_ids = ids0 + ids1 + ids2
        assert len(retrieved_ids) == len(set(retrieved_ids))
        assert sorted(retrieved_ids) == expected_ids

        # rewind behaviour
        cur2.rewind(initial_page_state="some string")
        cur2.rewind()
        with pytest.raises(ValueError, match="null"):
            cur2.rewind(initial_page_state=None)  # type: ignore[arg-type]

    @pytest.mark.describe("test of collection cursors, fetch_next_page, async")
    async def test_collection_cursors_fetchnextpage_async(
        self,
        async_filled_pagination_collection: DefaultAsyncCollection,
    ) -> None:
        cur0 = async_filled_pagination_collection.find(filter={"even": True})
        page0 = await cur0.fetch_next_page()
        ids0 = [doc["_id"] for doc in page0.results]
        nps0 = page0.next_page_state
        assert isinstance(nps0, str)

        cur1 = async_filled_pagination_collection.find(
            filter={"even": True},
            initial_page_state=nps0,
        )
        page1 = await cur1.fetch_next_page()
        ids1 = [doc["_id"] for doc in page1.results]
        nps1 = page1.next_page_state
        assert isinstance(nps1, str)

        cur2 = async_filled_pagination_collection.find(
            filter={"even": True},
            initial_page_state=nps1,
        )
        page2 = await cur2.fetch_next_page()
        ids2 = [doc["_id"] for doc in page2.results]
        assert page2.next_page_state is None

        expected_ids = [i for i in range(NUM_DOCS_PAGINATION) if i % 2 == 0]
        retrieved_ids = ids0 + ids1 + ids2
        assert len(retrieved_ids) == len(set(retrieved_ids))
        assert sorted(retrieved_ids) == expected_ids

        # fetching consecutive pages on a given cursor
        cur0x = async_filled_pagination_collection.find(filter={"even": True})
        await cur0x.fetch_next_page()
        page1x = await cur0x.fetch_next_page()
        assert page1x == page1

        # forbidden: mixing pagination and ordinary usage
        await cur0x.__anext__()
        with pytest.raises(CursorException):
            await cur0x.fetch_next_page()
        await cur0x.to_list()
        with pytest.raises(CursorException):
            await cur0x.fetch_next_page()

        # mapping
        cur0y = async_filled_pagination_collection.find(filter={"even": True}).map(
            lambda doc: doc["_id"]
        )
        page0y = await cur0y.fetch_next_page()
        assert page0y.results == ids0

        # vector ANN one-page behaviour re: include_sort_vector and its format
        coll_noncustom = async_filled_pagination_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=False),
            )
        )
        coll_custom = async_filled_pagination_collection.with_options(
            api_options=APIOptions(
                serdes_options=SerdesOptions(custom_datatypes_in_reading=True),
            )
        )

        vcur0_nc = coll_noncustom.find(
            sort={"$vector": [1, 1]},
            include_sort_vector=True,
            limit=15,
        )
        vpage0_nc = await vcur0_nc.fetch_next_page()
        assert vpage0_nc.next_page_state is None
        assert len(vpage0_nc.results) == 15
        assert isinstance(vpage0_nc.sort_vector, list)
        assert not isinstance(vpage0_nc.sort_vector, DataAPIVector)

        vcur0_c = coll_custom.find(
            sort={"$vector": [1, 1]},
            include_sort_vector=True,
            limit=15,
        )
        vpage0_c = await vcur0_c.fetch_next_page()
        assert vpage0_c.next_page_state is None
        assert len(vpage0_c.results) == 15
        assert isinstance(vpage0_c.sort_vector, DataAPIVector)



from __future__ import annotations

import time

import pytest
import werkzeug
from pytest_httpserver import HTTPServer

from astrapy import Collection, Database
from astrapy.exceptions import DataAPITimeoutException
from astrapy.utils.api_options import defaultAPIOptions
from astrapy.utils.request_tools import HttpMethod

from ..conftest import DefaultAsyncCollection, DefaultCollection

SLEEPER_TIME_MS = 500
TIMEOUT_PARAM_MS = 1

BASE_PATH = "v1"
KEYSPACE = "keyspace"
COLLECTION_NAME = "collection"
PATH_SUFFIX = f"{KEYSPACE}/{COLLECTION_NAME}"


@pytest.fixture
def mock_collection(httpserver: HTTPServer) -> DefaultCollection:
    base_endpoint = httpserver.url_for("/")
    coll: DefaultCollection = Collection(
        database=Database(
            api_endpoint=base_endpoint,
            keyspace=KEYSPACE,
            api_options=defaultAPIOptions(environment="other"),
        ),
        name=COLLECTION_NAME,
        keyspace=None,
        api_options=defaultAPIOptions(environment="other"),
    )
    return coll


@pytest.fixture
def mock_acollection(
    httpserver: HTTPServer, mock_collection: DefaultCollection
) -> DefaultAsyncCollection:
    return mock_collection.to_async()


def response_sleeper(request: werkzeug.Request) -> werkzeug.Response:
    time.sleep(SLEEPER_TIME_MS / 1000)
    return werkzeug.Response()


@pytest.mark.skip("Suppressed. The HTTPServer plays tricks here sometimes")
class TestCollectionTimeouts:
    @pytest.mark.describe("test of collection count_documents timeout, sync")
    def test_collection_count_documents_timeout_sync(
        self,
        httpserver: HTTPServer,
        mock_collection: DefaultCollection,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 500}})
        mock_collection.count_documents({}, upper_bound=800)

        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DataAPITimeoutException):
            mock_collection.count_documents(
                {}, upper_bound=800, general_method_timeout_ms=TIMEOUT_PARAM_MS
            )

    @pytest.mark.describe("test of collection count_documents timeout, async")
    async def test_collection_count_documents_timeout_async(
        self,
        httpserver: HTTPServer,
        mock_acollection: DefaultAsyncCollection,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"status": {"count": 500}})
        await mock_acollection.count_documents({}, upper_bound=800)

        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DataAPITimeoutException):
            await mock_acollection.count_documents(
                {}, upper_bound=800, general_method_timeout_ms=TIMEOUT_PARAM_MS
            )

    @pytest.mark.describe("test of collection cursor-based timeouts, async")
    async def test_collection_cursor_timeouts_async(
        self,
        httpserver: HTTPServer,
        mock_acollection: DefaultAsyncCollection,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"data": {"nextPageState": None, "documents": [{"a": 1}]}})
        cur0 = mock_acollection.find({})
        await cur0.__anext__()

        cur1 = mock_acollection.find({}, request_timeout_ms=1)
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DataAPITimeoutException):
            await cur1.__anext__()

    @pytest.mark.describe("test of collection find_one timeouts, async")
    async def test_collection_find_one_timeouts_async(
        self,
        httpserver: HTTPServer,
        mock_acollection: DefaultAsyncCollection,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"data": {"document": {"a": 1}}})
        await mock_acollection.find_one({})

        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DataAPITimeoutException):
            await mock_acollection.find_one({}, general_method_timeout_ms=1)

    @pytest.mark.describe("test of collection cursor-based timeouts, sync")
    def test_collection_cursor_timeouts_sync(
        self,
        httpserver: HTTPServer,
        mock_collection: DefaultCollection,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"data": {"nextPageState": None, "documents": [{"a": 1}]}})
        cur0 = mock_collection.find({})
        cur0.__next__()

        cur1 = mock_collection.find({}, request_timeout_ms=1)
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DataAPITimeoutException):
            cur1.__next__()

    @pytest.mark.describe("test of collection find_one timeouts, sync")
    def test_collection_find_one_timeouts_sync(
        self,
        httpserver: HTTPServer,
        mock_collection: DefaultCollection,
    ) -> None:
        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_json({"data": {"document": [{"a": 1}]}})
        mock_collection.find_one({})

        httpserver.expect_oneshot_request(
            f"/{BASE_PATH}/{PATH_SUFFIX}",
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DataAPITimeoutException):
            mock_collection.find_one({}, general_method_timeout_ms=1)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the validation/parsing of collection options
"""

from __future__ import annotations

from typing import Any

import pytest

from astrapy.info import CollectionDefinition, CollectionDescriptor


@pytest.mark.describe(
    "test of recasting the collection definition options from the api"
)
def test_recast_api_collection_dict() -> None:
    api_coll_descs: list[dict[str, Any]] = [
        # minimal:
        {
            "name": "col_name",
        },
        # full, w/o service:
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "metric": "cosine",
                    "service": {
                        "provider": "nvidia",
                        "modelName": "nvidia/nv-embedqa-e5-v5",
                    },
                },
                "indexing": {"deny": ["a"]},
                "defaultId": {"type": "objectId"},
            },
        },
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "metric": "cosine",
                },
                "indexing": {"deny": ["a"]},
                "defaultId": {"type": "objectId"},
            },
        },
        # partial/absent "vector", w/o service:
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "metric": "cosine",
                },
                "indexing": {"deny": ["a"]},
                "defaultId": {"type": "objectId"},
            },
        },
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "metric": "cosine",
                    "service": {
                        "provider": "nvidia",
                        "modelName": "nvidia/nv-embedqa-e5-v5",
                    },
                },
                "indexing": {"deny": ["a"]},
                "defaultId": {"type": "objectId"},
            },
        },
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                },
                "indexing": {"deny": ["a"]},
                "defaultId": {"type": "objectId"},
            },
        },
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "service": {
                        "provider": "nvidia",
                        "modelName": "nvidia/nv-embedqa-e5-v5",
                    },
                },
                "indexing": {"deny": ["a"]},
                "defaultId": {"type": "objectId"},
            },
        },
        {
            "name": "col_name",
            "options": {
                "indexing": {"deny": ["a"]},
                "defaultId": {"type": "objectId"},
            },
        },
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "service": {
                        "provider": "nvidia",
                        "modelName": "nvidia/nv-embedqa-e5-v5",
                    },
                },
                "indexing": {"deny": ["a"]},
                "defaultId": {"type": "objectId"},
            },
        },
        {
            "name": "col_name",
            "options": {
                "indexing": {"deny": ["a"]},
                "defaultId": {"type": "objectId"},
            },
        },
        # no indexing:
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "metric": "cosine",
                },
                "defaultId": {"type": "objectId"},
            },
        },
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "metric": "cosine",
                    "service": {
                        "provider": "nvidia",
                        "modelName": "nvidia/nv-embedqa-e5-v5",
                    },
                },
                "defaultId": {"type": "objectId"},
            },
        },
        # no defaultId:
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "metric": "cosine",
                },
                "indexing": {"deny": ["a"]},
            },
        },
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "metric": "cosine",
                    "service": {
                        "provider": "nvidia",
                        "modelName": "nvidia/nv-embedqa-e5-v5",
                    },
                },
                "indexing": {"deny": ["a"]},
            },
        },
        # no indexing + no defaultId:
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "metric": "cosine",
                },
            },
        },
        {
            "name": "col_name",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "metric": "cosine",
                    "service": {
                        "provider": "nvidia",
                        "modelName": "nvidia/nv-embedqa-e5-v5",
                    },
                },
            },
        },
        # a real one (full)
        {
            "name": "testcoll",
            "options": {
                "vector": {
                    "dimension": 1024,
                    "metric": "cosine",
                    "service": {
                        "provider": "voyageAI",
                        "modelName": "voyage-large-2-instruct",
                        "authentication": {
                            "providerKey": "SHARED_SECRET_EMBEDDING_API_KEY_VOYAGEAI.providerKey",
                        },
                        "parameters": {
                            "autoTruncate": False,
                        },
                    },
                    "sourceModel": "other",
                },
                "indexing": {
                    "allow": [
                        "xing",
                    ],
                },
                "defaultId": {
                    "type": "objectId",
                },
            },
        },
    ]
    for cd_dict in api_coll_descs:
        descriptor = CollectionDescriptor._from_dict(cd_dict)
        # dict->obj->dict test
        assert descriptor.as_dict() == cd_dict
        if "options" in cd_dict:
            cdef_dict = cd_dict["options"]
            assert CollectionDefinition._from_dict(cdef_dict).as_dict() == cdef_dict
        # obj->dict->obj test
        assert CollectionDescriptor._from_dict(descriptor.as_dict()) == descriptor
        definition = descriptor.definition
        assert CollectionDefinition._from_dict(definition.as_dict()) == definition
        # coerce calls
        assert CollectionDescriptor.coerce(cd_dict) == descriptor
        assert CollectionDescriptor.coerce(descriptor) == descriptor
        if "options" in cd_dict:
            assert CollectionDefinition.coerce(cd_dict["options"]) == definition
        assert CollectionDefinition.coerce(definition) == definition


@pytest.mark.describe("test of fluent interface for CollectionDefinition")
def test_fluent_collection_definition() -> None:
    zero = CollectionDefinition.builder()
    assert zero.build().as_dict() == {}

    rich = (
        zero.set_indexing("allow", ["a", "b"])
        .set_default_id("UUID")
        .set_vector_dimension(123)
        .set_vector_metric("cosine")
        .set_vector_service(
            "prov", "mod", authentication={"a": "u"}, parameters={"p": "a"}
        )
        .build()
    )
    expected_rich_dict = {
        "vector": {
            "dimension": 123,
            "metric": "cosine",
            "service": {
                "provider": "prov",
                "modelName": "mod",
                "authentication": {"a": "u"},
                "parameters": {"p": "a"},
            },
        },
        "indexing": {"allow": ["a", "b"]},
        "defaultId": {"type": "UUID"},
    }
    assert rich.as_dict() == expected_rich_dict

    zero_2 = (
        rich.set_indexing(None)
        .set_default_id(None)
        .set_vector_dimension(None)
        .set_vector_metric(None)
        .set_vector_source_model(None)
        .set_vector_service(None)
        .build()
    )
    assert zero_2.build().as_dict() == {}



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from astrapy.utils.str_enum import StrEnum


class Environment:
    """
    Admitted values for `environment` property,
    denoting the targeted API deployment type.
    """

    def __init__(self) -> None:
        raise NotImplementedError

    PROD = "prod"
    DEV = "dev"
    TEST = "test"
    DSE = "dse"
    HCD = "hcd"
    CASSANDRA = "cassandra"
    OTHER = "other"

    values = {PROD, DEV, TEST, DSE, HCD, CASSANDRA, OTHER}
    astra_db_values = {PROD, DEV, TEST}


class ModelStatus(StrEnum):
    """
    Admitted values for the status of a (reranking/embedding) model,
    as returned by the corresponding Data API query.
    """

    ALL = ""
    SUPPORTED = "SUPPORTED"
    DEPRECATED = "DEPRECATED"
    END_OF_LIFE = "END_OF_LIFE"


class DatabaseStatus(StrEnum):
    """
    Admitted values for the status of a database as returned by Astra DB.

    To avoid the risk of a newly-introduced status (by the DevOps API)
    breaking the clients, this enum is used only for reference and no coercion
    from server-received status strings is enforced.
    """

    ACTIVE = "ACTIVE"
    ASSOCIATING = "ASSOCIATING"
    DECOMMISSIONING = "DECOMMISSIONING"
    DEGRADED = "DEGRADED"
    ERROR = "ERROR"
    HIBERNATED = "HIBERNATED"
    HIBERNATING = "HIBERNATING"
    INITIALIZING = "INITIALIZING"
    MAINTENANCE = "MAINTENANCE"
    PARKED = "PARKED"
    PARKING = "PARKING"
    PENDING = "PENDING"
    PREPARED = "PREPARED"
    PREPARING = "PREPARING"
    RESIZING = "RESIZING"
    RESUMING = "RESUMING"
    SYNCHRONIZING = "SYNCHRONIZING"
    TERMINATED = "TERMINATED"
    TERMINATING = "TERMINATING"
    UNKNOWN = "UNKNOWN"
    UNPARKING = "UNPARKING"



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.data.utils.table_converters import create_row_tpostprocessor
from astrapy.data_types import (
    DataAPIDictUDT,
    DataAPIMap,
    DataAPISet,
)
from astrapy.info import ListTableDescriptor
from astrapy.utils.api_options import SerdesOptions, defaultSerdesOptions

from ..table_structure_assets import dict_equal_same_class
from ..table_udt_assets import (
    THE_BYTES,
    THE_DATETIME,
    THE_SERIALIZED_BYTES,
    THE_SERIALIZED_TIMESTAMP,
    THE_TIMESTAMP,
    THE_TIMEZONE,
    UNIT_EXTENDED_PLAYER_TYPE_DEFINITION,
    UNIT_EXTENDED_PLAYER_TYPE_NAME,
    UNIT_OPTLST_PLAYER_TYPE_DEFINITION,
    UNIT_OPTLST_PLAYER_TYPE_NAME,
    UnitExtendedPlayer,
    UnitNullableRequiringPlayer,
    _unit_extended_player_from_dict,
    _unit_optlst_player_from_dict,
)

TABLE_DESCRIPTION = {
    "name": "table_unit_udt_deserialize_test",
    "definition": {
        "columns": {
            "p_text": {"type": "text"},
            "scalar_udt": {
                "type": "userDefined",
                "udtName": UNIT_EXTENDED_PLAYER_TYPE_NAME,
                "definition": UNIT_EXTENDED_PLAYER_TYPE_DEFINITION.as_dict(),
            },
            "list_udt": {
                "type": "list",
                "valueType": {
                    "type": "userDefined",
                    "udtName": UNIT_EXTENDED_PLAYER_TYPE_NAME,
                    "definition": UNIT_EXTENDED_PLAYER_TYPE_DEFINITION.as_dict(),
                },
            },
            "set_udt": {
                "type": "set",
                "valueType": {
                    "type": "userDefined",
                    "udtName": UNIT_EXTENDED_PLAYER_TYPE_NAME,
                    "definition": UNIT_EXTENDED_PLAYER_TYPE_DEFINITION.as_dict(),
                },
            },
            "map_str_udt": {
                "type": "map",
                "keyType": "text",
                "valueType": {
                    "type": "userDefined",
                    "udtName": UNIT_EXTENDED_PLAYER_TYPE_NAME,
                    "definition": UNIT_EXTENDED_PLAYER_TYPE_DEFINITION.as_dict(),
                },
            },
            "map_int_udt_aslist": {
                "type": "map",
                "keyType": "int",
                "valueType": {
                    "type": "userDefined",
                    "udtName": UNIT_EXTENDED_PLAYER_TYPE_NAME,
                    "definition": UNIT_EXTENDED_PLAYER_TYPE_DEFINITION.as_dict(),
                },
            },
        },
        "primaryKey": {"partitionBy": [], "partitionSort": {}},
    },
}
COLUMNS = ListTableDescriptor.coerce(TABLE_DESCRIPTION).definition.columns

INCOMPLETE_UDTS_TABLE_DESCRIPTION = {
    "name": "table_unit_incomplete_udt_deserialize_test",
    "definition": {
        "columns": {
            "p_text": {"type": "text"},
            "scalar_udt": {
                "type": "userDefined",
                "udtName": UNIT_OPTLST_PLAYER_TYPE_NAME,
                "definition": UNIT_OPTLST_PLAYER_TYPE_DEFINITION.as_dict(),
            },
            "list_udts": {
                "type": "list",
                "valueType": {
                    "type": "userDefined",
                    "udtName": UNIT_OPTLST_PLAYER_TYPE_NAME,
                    "definition": UNIT_OPTLST_PLAYER_TYPE_DEFINITION.as_dict(),
                },
            },
            "set_udts": {
                "type": "set",
                "valueType": {
                    "type": "userDefined",
                    "udtName": UNIT_OPTLST_PLAYER_TYPE_NAME,
                    "definition": UNIT_OPTLST_PLAYER_TYPE_DEFINITION.as_dict(),
                },
            },
            "map_udts": {
                "type": "map",
                "keyType": "int",
                "valueType": {
                    "type": "userDefined",
                    "udtName": UNIT_OPTLST_PLAYER_TYPE_NAME,
                    "definition": UNIT_OPTLST_PLAYER_TYPE_DEFINITION.as_dict(),
                },
            },
        },
        "primaryKey": {"partitionBy": ["p_text"], "partitionSort": {}},
    },
}
INCOMPLETE_UDTS_TABLE_COLUMNS = ListTableDescriptor.coerce(
    INCOMPLETE_UDTS_TABLE_DESCRIPTION,
).definition.columns
INCOMPLETE_UDTS_PARTIAL_UDT_INPUT = {"age": 42}
INCOMPLETE_UDTS_FULL_UDT_INPUT = {"name": "Kyle", "age": 18, "victories": ["x", "y"]}
INCOMPLETE_UDTS_NULL_UDT_RESULT = UnitNullableRequiringPlayer(
    name=None,
    age=None,
    victories=[],
)
INCOMPLETE_UDTS_PARTIAL_UDT_RESULT = UnitNullableRequiringPlayer(
    name=None,
    age=42,
    victories=[],
)
INCOMPLETE_UDTS_FULL_UDT_RESULT = UnitNullableRequiringPlayer(
    name="Kyle",
    age=18,
    victories=["x", "y"],
)
#
INCOMPLETE_UDTS_OMITTED_OUTPUT_ROW_TO_POSTPROCESS = {"p_text": "omitteds"}
INCOMPLETE_UDTS_NULLFILLED_OUTPUT_ROW_TO_POSTPROCESS = {
    "p_text": "nulls",
    "scalar_udt": None,
    "list_udts": [None],
    "set_udts": [None],
    "map_udts": {123: None},
}
INCOMPLETE_UDTS_NULLFILLED_OUTPUT_ROW_TO_POSTPROCESS_TUPLES = {
    "p_text": "nulls",
    "scalar_udt": None,
    "list_udts": [None],
    "set_udts": [None],
    "map_udts": [[123, None]],
}
INCOMPLETE_UDTS_EMPTIES_OUTPUT_ROW_TO_POSTPROCESS = {
    "p_text": "empties",
    "scalar_udt": {},
    "list_udts": [{}],
    "set_udts": [{}],
    "map_udts": {321: {}},
}
INCOMPLETE_UDTS_EMPTIES_OUTPUT_ROW_TO_POSTPROCESS_TUPLES = {
    "p_text": "empties",
    "scalar_udt": {},
    "list_udts": [{}],
    "set_udts": [{}],
    "map_udts": [[321, {}]],
}
INCOMPLETE_UDTS_PARTIAL_OUTPUT_ROW_TO_POSTPROCESS = {
    "p_text": "partials",
    "scalar_udt": INCOMPLETE_UDTS_PARTIAL_UDT_INPUT,
    "list_udts": [INCOMPLETE_UDTS_PARTIAL_UDT_INPUT],
    "set_udts": [INCOMPLETE_UDTS_PARTIAL_UDT_INPUT],
    "map_udts": {456: INCOMPLETE_UDTS_PARTIAL_UDT_INPUT},
}
INCOMPLETE_UDTS_PARTIAL_OUTPUT_ROW_TO_POSTPROCESS_TUPLES = {
    "p_text": "partials",
    "scalar_udt": INCOMPLETE_UDTS_PARTIAL_UDT_INPUT,
    "list_udts": [INCOMPLETE_UDTS_PARTIAL_UDT_INPUT],
    "set_udts": [INCOMPLETE_UDTS_PARTIAL_UDT_INPUT],
    "map_udts": [[456, INCOMPLETE_UDTS_PARTIAL_UDT_INPUT]],
}
INCOMPLETE_UDTS_FULL_OUTPUT_ROW_TO_POSTPROCESS = {
    "p_text": "fulls",
    "scalar_udt": INCOMPLETE_UDTS_FULL_UDT_INPUT,
    "list_udts": [INCOMPLETE_UDTS_FULL_UDT_INPUT],
    "set_udts": [INCOMPLETE_UDTS_FULL_UDT_INPUT],
    "map_udts": {789: INCOMPLETE_UDTS_FULL_UDT_INPUT},
}
INCOMPLETE_UDTS_FULL_OUTPUT_ROW_TO_POSTPROCESS_TUPLES = {
    "p_text": "fulls",
    "scalar_udt": INCOMPLETE_UDTS_FULL_UDT_INPUT,
    "list_udts": [INCOMPLETE_UDTS_FULL_UDT_INPUT],
    "set_udts": [INCOMPLETE_UDTS_FULL_UDT_INPUT],
    "map_udts": [[789, INCOMPLETE_UDTS_FULL_UDT_INPUT]],
}
#
INCOMPLETE_UDTS_OMITTED_OUTPUT_EXPECTED_ROW = {
    "p_text": "omitteds",
    "scalar_udt": INCOMPLETE_UDTS_NULL_UDT_RESULT,
    "list_udts": [],
    "set_udts": DataAPISet([]),
    "map_udts": DataAPIMap({}),
}
INCOMPLETE_UDTS_NULLFILLED_OUTPUT_EXPECTED_ROW = {
    "p_text": "nulls",
    "scalar_udt": INCOMPLETE_UDTS_NULL_UDT_RESULT,
    "list_udts": [INCOMPLETE_UDTS_NULL_UDT_RESULT],
    "set_udts": DataAPISet([INCOMPLETE_UDTS_NULL_UDT_RESULT]),
    "map_udts": DataAPIMap({123: INCOMPLETE_UDTS_NULL_UDT_RESULT}),
}
INCOMPLETE_UDTS_EMPTIES_OUTPUT_EXPECTED_ROW = {
    "p_text": "empties",
    "scalar_udt": INCOMPLETE_UDTS_NULL_UDT_RESULT,
    "list_udts": [INCOMPLETE_UDTS_NULL_UDT_RESULT],
    "set_udts": DataAPISet([INCOMPLETE_UDTS_NULL_UDT_RESULT]),
    "map_udts": DataAPIMap({321: INCOMPLETE_UDTS_NULL_UDT_RESULT}),
}
INCOMPLETE_UDTS_PARTIAL_OUTPUT_EXPECTED_ROW = {
    "p_text": "partials",
    "scalar_udt": INCOMPLETE_UDTS_PARTIAL_UDT_RESULT,
    "list_udts": [INCOMPLETE_UDTS_PARTIAL_UDT_RESULT],
    "set_udts": DataAPISet([INCOMPLETE_UDTS_PARTIAL_UDT_RESULT]),
    "map_udts": DataAPIMap({456: INCOMPLETE_UDTS_PARTIAL_UDT_RESULT}),
}
INCOMPLETE_UDTS_FULL_OUTPUT_EXPECTED_ROW = {
    "p_text": "fulls",
    "scalar_udt": INCOMPLETE_UDTS_FULL_UDT_RESULT,
    "list_udts": [INCOMPLETE_UDTS_FULL_UDT_RESULT],
    "set_udts": DataAPISet([INCOMPLETE_UDTS_FULL_UDT_RESULT]),
    "map_udts": DataAPIMap({789: INCOMPLETE_UDTS_FULL_UDT_RESULT}),
}

RAW_RESPONSE_UDT_DICT = {
    "name": "John",
    "age": 40,
    "blb": THE_SERIALIZED_BYTES,
    "ts": THE_SERIALIZED_TIMESTAMP,
}
EXPECTED_STDLIB_DICT = {
    "name": "John",
    "age": 40,
    "blb": THE_BYTES,
    "ts": THE_DATETIME,
}
EXPECTED_DICTUDT = DataAPIDictUDT(
    {
        "name": "John",
        "age": 40,
        "blb": THE_BYTES,
        "ts": THE_TIMESTAMP,
    }
)
EXPECTED_STDLIB_CUSTOMUDT = UnitExtendedPlayer(
    name="John",
    age=40,
    blb=THE_BYTES,
    ts=THE_DATETIME,
)
EXPECTED_CUSTOMUDT = UnitExtendedPlayer(
    name="John",
    age=40,
    blb=THE_BYTES,
    ts=THE_TIMESTAMP,
)

OUTPUT_ROW_TO_POSTPROCESS = {
    "p_text": "base",
    "scalar_udt": RAW_RESPONSE_UDT_DICT,
    "list_udt": [RAW_RESPONSE_UDT_DICT],
    "set_udt": [RAW_RESPONSE_UDT_DICT],
    "map_str_udt": {"k": RAW_RESPONSE_UDT_DICT},
    "map_int_udt_aslist": [[101, RAW_RESPONSE_UDT_DICT]],
}

EXPECTED_ROW_STDLIB_DICT = {
    "p_text": "base",
    "scalar_udt": EXPECTED_STDLIB_DICT,
    "list_udt": [EXPECTED_STDLIB_DICT],
    "set_udt": "ignored-for-this-test",
    "map_str_udt": {"k": EXPECTED_STDLIB_DICT},
    "map_int_udt_aslist": {101: EXPECTED_STDLIB_DICT},
}
EXPECTED_ROW_DICTUDT = {
    "p_text": "base",
    "scalar_udt": EXPECTED_DICTUDT,
    "list_udt": [EXPECTED_DICTUDT],
    "set_udt": DataAPISet([EXPECTED_DICTUDT]),
    "map_str_udt": DataAPIMap([("k", EXPECTED_DICTUDT)]),
    "map_int_udt_aslist": DataAPIMap([(101, EXPECTED_DICTUDT)]),
}
EXPECTED_ROW_STDLIB_CUSTOMUDT = {
    "p_text": "base",
    "scalar_udt": EXPECTED_STDLIB_CUSTOMUDT,
    "list_udt": [EXPECTED_STDLIB_CUSTOMUDT],
    "set_udt": "ignored-for-this-test",
    "map_str_udt": {"k": EXPECTED_STDLIB_CUSTOMUDT},
    "map_int_udt_aslist": {101: EXPECTED_STDLIB_CUSTOMUDT},
}
EXPECTED_ROW_CUSTOMUDT = {
    "p_text": "base",
    "scalar_udt": EXPECTED_CUSTOMUDT,
    "list_udt": [EXPECTED_CUSTOMUDT],
    "set_udt": DataAPISet([EXPECTED_CUSTOMUDT]),
    "map_str_udt": DataAPIMap([("k", EXPECTED_CUSTOMUDT)]),
    "map_int_udt_aslist": DataAPIMap([(101, EXPECTED_CUSTOMUDT)]),
}

BASE_OPTIONS = defaultSerdesOptions.with_override(
    SerdesOptions(datetime_tzinfo=THE_TIMEZONE),
)
OPTIONS_STDLIB = BASE_OPTIONS.with_override(
    SerdesOptions(custom_datatypes_in_reading=False),
)
OPTIONS_CUSTOM = BASE_OPTIONS.with_override(
    SerdesOptions(custom_datatypes_in_reading=True),
)
OPTIONS_STDLIB_CCLASS = OPTIONS_STDLIB.with_override(
    SerdesOptions(
        deserializer_by_udt={
            UNIT_EXTENDED_PLAYER_TYPE_NAME: _unit_extended_player_from_dict,
        }
    ),
)
OPTIONS_CUSTOM_CCLASS = OPTIONS_CUSTOM.with_override(
    SerdesOptions(
        deserializer_by_udt={
            UNIT_EXTENDED_PLAYER_TYPE_NAME: _unit_extended_player_from_dict,
        }
    ),
)
INCOMPLETE_DESERIALIZER_OPTIONS = OPTIONS_CUSTOM.with_override(
    SerdesOptions(
        deserializer_by_udt={
            UNIT_OPTLST_PLAYER_TYPE_NAME: _unit_optlst_player_from_dict,
        }
    ),
)


class TestTPostProcessorsUserDefinedTypes:
    @pytest.mark.describe(
        "test of UDTs in row postprocessors: custom datatypes, custom class"
    )
    def test_row_postprocessors_udts_customdt_customclass(self) -> None:
        tpostprocessor = create_row_tpostprocessor(
            columns=COLUMNS,
            options=OPTIONS_CUSTOM_CCLASS,
            similarity_pseudocolumn=None,
        )
        deserialized_row = tpostprocessor(OUTPUT_ROW_TO_POSTPROCESS)
        dict_equal_same_class(deserialized_row, EXPECTED_ROW_CUSTOMUDT)

    @pytest.mark.describe(
        "test of UDTs in row postprocessors: stdlib datatypes, custom class"
    )
    def test_row_postprocessors_udts_stdlibdt_customclass(self) -> None:
        tpostprocessor = create_row_tpostprocessor(
            columns=COLUMNS,
            options=OPTIONS_STDLIB_CCLASS,
            similarity_pseudocolumn=None,
        )
        # remove set column due to non-hashability
        deserialized_row = tpostprocessor(
            {k: v for k, v in OUTPUT_ROW_TO_POSTPROCESS.items() if "set" not in k}
        )
        expected_row = {
            **EXPECTED_ROW_STDLIB_CUSTOMUDT,
            **{"set_udt": set()},
        }
        dict_equal_same_class(deserialized_row, expected_row)

    @pytest.mark.describe(
        "test of UDTs in row postprocessors: custom datatypes, dict-wrapper class"
    )
    def test_row_postprocessors_udts_customdt_dictwrapperclass(self) -> None:
        tpostprocessor = create_row_tpostprocessor(
            columns=COLUMNS,
            options=OPTIONS_CUSTOM,
            similarity_pseudocolumn=None,
        )
        deserialized_row = tpostprocessor(OUTPUT_ROW_TO_POSTPROCESS)
        dict_equal_same_class(deserialized_row, EXPECTED_ROW_DICTUDT)

    @pytest.mark.describe(
        "test of UDTs in row postprocessors: stdlib datatypes, dict-wrapper class"
    )
    def test_row_postprocessors_udts_stdlibdt_dictwrapperclass(self) -> None:
        tpostprocessor = create_row_tpostprocessor(
            columns=COLUMNS,
            options=OPTIONS_STDLIB,
            similarity_pseudocolumn=None,
        )
        # remove set column due to non-hashability
        deserialized_row = tpostprocessor(
            {k: v for k, v in OUTPUT_ROW_TO_POSTPROCESS.items() if "set" not in k}
        )
        expected_row = {
            **EXPECTED_ROW_STDLIB_DICT,
            **{"set_udt": set()},
        }
        dict_equal_same_class(deserialized_row, expected_row)

    @pytest.mark.describe("test of row postprocessors, incomplete UDT: omitted data")
    def test_row_postprocessors_incomplete_omitted_udts(self) -> None:
        tpostprocessor = create_row_tpostprocessor(
            columns=INCOMPLETE_UDTS_TABLE_COLUMNS,
            options=INCOMPLETE_DESERIALIZER_OPTIONS,
            similarity_pseudocolumn=None,
        )
        deserialized_row = tpostprocessor(
            INCOMPLETE_UDTS_OMITTED_OUTPUT_ROW_TO_POSTPROCESS,
        )
        dict_equal_same_class(
            deserialized_row,
            INCOMPLETE_UDTS_OMITTED_OUTPUT_EXPECTED_ROW,
        )

    @pytest.mark.describe("test of row postprocessors, incomplete UDT: nullfilled data")
    def test_row_postprocessors_incomplete_nullfilled_udts(self) -> None:
        tpostprocessor = create_row_tpostprocessor(
            columns=INCOMPLETE_UDTS_TABLE_COLUMNS,
            options=INCOMPLETE_DESERIALIZER_OPTIONS,
            similarity_pseudocolumn=None,
        )
        deserialized_row = tpostprocessor(
            INCOMPLETE_UDTS_NULLFILLED_OUTPUT_ROW_TO_POSTPROCESS,
        )
        dict_equal_same_class(
            deserialized_row,
            INCOMPLETE_UDTS_NULLFILLED_OUTPUT_EXPECTED_ROW,
        )
        deserialized_row_tuples = tpostprocessor(
            INCOMPLETE_UDTS_NULLFILLED_OUTPUT_ROW_TO_POSTPROCESS_TUPLES,
        )
        dict_equal_same_class(
            deserialized_row_tuples,
            INCOMPLETE_UDTS_NULLFILLED_OUTPUT_EXPECTED_ROW,
        )

    @pytest.mark.describe("test of row postprocessors, incomplete UDT: emptydict data")
    def test_row_postprocessors_incomplete_emptydict_udts(self) -> None:
        tpostprocessor = create_row_tpostprocessor(
            columns=INCOMPLETE_UDTS_TABLE_COLUMNS,
            options=INCOMPLETE_DESERIALIZER_OPTIONS,
            similarity_pseudocolumn=None,
        )
        deserialized_row = tpostprocessor(
            INCOMPLETE_UDTS_EMPTIES_OUTPUT_ROW_TO_POSTPROCESS,
        )
        dict_equal_same_class(
            deserialized_row,
            INCOMPLETE_UDTS_EMPTIES_OUTPUT_EXPECTED_ROW,
        )
        deserialized_row_tuples = tpostprocessor(
            INCOMPLETE_UDTS_EMPTIES_OUTPUT_ROW_TO_POSTPROCESS_TUPLES,
        )
        dict_equal_same_class(
            deserialized_row_tuples,
            INCOMPLETE_UDTS_EMPTIES_OUTPUT_EXPECTED_ROW,
        )

    @pytest.mark.describe("test of row postprocessors, incomplete UDT: partial data")
    def test_row_postprocessors_incomplete_partial_udts(self) -> None:
        tpostprocessor = create_row_tpostprocessor(
            columns=INCOMPLETE_UDTS_TABLE_COLUMNS,
            options=INCOMPLETE_DESERIALIZER_OPTIONS,
            similarity_pseudocolumn=None,
        )
        deserialized_row = tpostprocessor(
            INCOMPLETE_UDTS_PARTIAL_OUTPUT_ROW_TO_POSTPROCESS,
        )
        dict_equal_same_class(
            deserialized_row,
            INCOMPLETE_UDTS_PARTIAL_OUTPUT_EXPECTED_ROW,
        )
        deserialized_row_tuples = tpostprocessor(
            INCOMPLETE_UDTS_PARTIAL_OUTPUT_ROW_TO_POSTPROCESS_TUPLES,
        )
        dict_equal_same_class(
            deserialized_row_tuples,
            INCOMPLETE_UDTS_PARTIAL_OUTPUT_EXPECTED_ROW,
        )

    @pytest.mark.describe("test of row postprocessors, incomplete UDT: full data")
    def test_row_postprocessors_incomplete_full_udts(self) -> None:
        tpostprocessor = create_row_tpostprocessor(
            columns=INCOMPLETE_UDTS_TABLE_COLUMNS,
            options=INCOMPLETE_DESERIALIZER_OPTIONS,
            similarity_pseudocolumn=None,
        )
        deserialized_row = tpostprocessor(
            INCOMPLETE_UDTS_FULL_OUTPUT_ROW_TO_POSTPROCESS,
        )
        dict_equal_same_class(
            deserialized_row,
            INCOMPLETE_UDTS_FULL_OUTPUT_EXPECTED_ROW,
        )
        deserialized_row_tuples = tpostprocessor(
            INCOMPLETE_UDTS_FULL_OUTPUT_ROW_TO_POSTPROCESS_TUPLES,
        )
        dict_equal_same_class(
            deserialized_row_tuples,
            INCOMPLETE_UDTS_FULL_OUTPUT_EXPECTED_ROW,
        )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the token providers
"""

from __future__ import annotations

import pytest

from astrapy.authentication import (
    StaticTokenProvider,
    UsernamePasswordTokenProvider,
    coerce_token_provider,
)


@pytest.mark.describe("test of static token provider")
def test_static_token_provider() -> None:
    literal_t = "AstraCS:1"
    static_tp = StaticTokenProvider(literal_t)

    assert static_tp.get_token() == literal_t


@pytest.mark.describe("test of username-password token provider")
def test_username_password_token_provider() -> None:
    up_tp = UsernamePasswordTokenProvider("cassandraA", "cassandraB")

    assert up_tp.get_token() == "Cassandra:Y2Fzc2FuZHJhQQ==:Y2Fzc2FuZHJhQg=="


@pytest.mark.describe("test of null token provider")
def test_null_token_provider() -> None:
    null_tp = StaticTokenProvider(None)

    assert null_tp.get_token() is None


@pytest.mark.describe("test of token providers coercion")
def test_coerce_token_provider() -> None:
    literal_t = "AstraCS:1"
    static_tp = StaticTokenProvider(literal_t)
    null_tp = StaticTokenProvider(None)
    up_tp = UsernamePasswordTokenProvider("cassandraA", "cassandraB")

    assert coerce_token_provider(literal_t).get_token() == literal_t
    assert coerce_token_provider(static_tp).get_token() == static_tp.get_token()
    assert coerce_token_provider(up_tp).get_token() == up_tp.get_token()
    assert coerce_token_provider(null_tp).get_token() is None


@pytest.mark.describe("test of token providers equality")
def test_token_provider_equality() -> None:
    literal_t = "AstraCS:1"
    static_tp_1 = StaticTokenProvider(literal_t)
    null_tp_1 = StaticTokenProvider(None)
    up_tp_1 = UsernamePasswordTokenProvider("cassandraA", "cassandraB")
    static_tp_2 = StaticTokenProvider(literal_t)
    null_tp_2 = StaticTokenProvider(None)
    up_tp_2 = UsernamePasswordTokenProvider("cassandraA", "cassandraB")

    assert static_tp_1 == static_tp_2
    assert null_tp_1 == null_tp_2
    assert up_tp_1 == up_tp_2

    assert static_tp_1 != null_tp_1
    assert static_tp_1 != up_tp_1
    assert null_tp_1 != static_tp_1
    assert null_tp_1 != up_tp_1
    assert up_tp_1 != static_tp_1
    assert up_tp_1 != null_tp_1

    # effective equality
    assert up_tp_1 == StaticTokenProvider("Cassandra:Y2Fzc2FuZHJhQQ==:Y2Fzc2FuZHJhQg==")


@pytest.mark.describe("test of token provider inheritance yield")
def test_token_provider_inheritance_yield() -> None:
    static_tp = StaticTokenProvider("AstraCS:xyz")
    null_tp = StaticTokenProvider(None)
    up_tp = UsernamePasswordTokenProvider("cassandraA", "cassandraB")

    assert static_tp | static_tp == static_tp
    assert static_tp | null_tp == static_tp
    assert static_tp | up_tp == static_tp

    assert null_tp | static_tp == static_tp
    assert null_tp | null_tp == null_tp
    assert null_tp | up_tp == up_tp

    assert up_tp | static_tp == up_tp
    assert up_tp | null_tp == up_tp
    assert up_tp | up_tp == up_tp

    assert static_tp or static_tp == static_tp
    assert static_tp or null_tp == static_tp
    assert static_tp or up_tp == static_tp

    assert null_tp or static_tp == static_tp
    assert null_tp or null_tp == null_tp
    assert null_tp or up_tp == up_tp

    assert up_tp or static_tp == up_tp
    assert up_tp or null_tp == up_tp
    assert up_tp or up_tp == up_tp



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import warnings
from functools import wraps
from typing import Callable, TypeVar

from deprecation import DeprecatedWarning
from typing_extensions import ParamSpec  # compatible with pre-3.10 Python

P = ParamSpec("P")
R = TypeVar("R")

BETA_WARNING_TEMPLATE = (
    "Method '{method_name}' is in beta and might undergo signature "
    "or behaviour changes in the future."
)


class BetaFeatureWarning(UserWarning):
    pass


def check_deprecated_alias(
    new_name: str | None,
    deprecated_name: str | None,
) -> str | None:
    """Generic blueprint utility for deprecating parameters through an alias.

    Normalize the two aliased parameter names, raising deprecation
    when needed and an error if both parameter supplied.
    The returned value is the final one for the parameter.
    """

    if deprecated_name is None:
        # no need for deprecation nor exceptions
        return new_name
    else:
        # issue a deprecation warning
        the_warning = DeprecatedWarning(
            "Parameter 'deprecated_name'",
            deprecated_in="2.0.0",
            removed_in="3.0.0",
            details="Please use 'new_name' instead.",
        )
        warnings.warn(
            the_warning,
            stacklevel=3,
        )

        if new_name is None:
            return deprecated_name
        else:
            msg = (
                "Parameters `new_name` and `deprecated_name` "
                "(a deprecated alias for the former) cannot be passed at the same time."
            )
            raise ValueError(msg)


def deprecated_property(
    new_name: str, deprecated_in: str, removed_in: str
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """
    Decorator for a @property that is a deprecated alias for attribute 'new_name'.
    """

    def _deprecator(method: Callable[P, R]) -> Callable[P, R]:
        @wraps(method)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            the_warning = DeprecatedWarning(
                f"Property '{method.__name__}'",
                deprecated_in=deprecated_in,
                removed_in=removed_in,
                details=f"Please use '{new_name}' instead.",
            )
            warnings.warn(
                the_warning,
                stacklevel=2,
            )
            return method(*args, **kwargs)

        return wrapper

    return _deprecator


def beta_method(method: Callable[P, R]) -> Callable[P, R]:
    @wraps(method)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        beta_warning_message = BETA_WARNING_TEMPLATE.format(
            method_name=method.__qualname__
        )
        warnings.warn(
            beta_warning_message,
            BetaFeatureWarning,
            stacklevel=2,
        )
        return method(*args, **kwargs)

    return wrapper


def issue_plain_warning(message: str, stacklevel: int = 2) -> None:
    warnings.warn(
        message,
        stacklevel=stacklevel,
    )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Iterable
from typing import TypedDict

import pytest

from astrapy import AsyncDatabase, AsyncTable, Database, Table
from astrapy.info import AlterTableAddColumns, AlterTableDropColumns

from ..conftest import DefaultAsyncTable, DefaultTable


class MyTestDoc(TypedDict):
    p_bigint: int
    p_ascii: str


class AlteredMyTestDoc(TypedDict):
    p_bigint: int
    p_ascii: str
    added: int


class MyTestMiniDoc(TypedDict):
    p_bigint: int


FIND_PROJECTION = {"p_bigint": True}
TYPING_TABLE_NAME = "test_typing_table"
TYPING_TABLE_DEFINITION = {
    "columns": {
        "p_ascii": "text",
        "p_bigint": "bigint",
        "p_float": "float",
    },
    "primaryKey": {
        "partitionBy": ["p_ascii", "p_bigint"],
        "partitionSort": {},
    },
}
ROW = {"p_ascii": "abc", "p_bigint": 10000, "p_float": 0.123}
TYPED_ROW: MyTestDoc = {"p_ascii": "abc", "p_bigint": 10000}
FIND_FILTER = {"p_ascii": "abc", "p_bigint": 10000}

ALTER_ADD_COLUMN = AlterTableAddColumns.coerce({"columns": {"added": "int"}})
ALTER_DROP_COLUMN = AlterTableDropColumns.coerce({"columns": ["added"]})


@pytest.fixture(scope="module")
def sync_typing_test_table(sync_database: Database) -> Iterable[DefaultTable]:
    table = sync_database.create_table(
        TYPING_TABLE_NAME,
        definition=TYPING_TABLE_DEFINITION,
        if_not_exists=True,
    )
    yield table

    table.drop()


@pytest.fixture
def async_typing_test_table(
    sync_typing_test_table: DefaultTable,
) -> DefaultAsyncTable:
    return sync_typing_test_table.to_async()


class TestTableTyping:
    @pytest.mark.describe("test of typing create_table and alter_table, sync")
    def test_create_alter_table_typing_sync(
        self,
        sync_database: Database,
        sync_typing_test_table: DefaultTable,
    ) -> None:
        """Test of creating typed tables with generics (and not), sync."""

        # Untyped baseline
        c_tb_untyped = sync_database.create_table(
            TYPING_TABLE_NAME,
            definition=TYPING_TABLE_DEFINITION,
            if_not_exists=True,
        )
        c_tb_untyped.insert_one(ROW)
        cu_doc = c_tb_untyped.find_one(FIND_FILTER)
        assert cu_doc is not None
        cu_a: str
        cu_b: int
        cu_a = cu_doc["p_ascii"]  # noqa: F841
        cu_b = cu_doc["p_bigint"]  # noqa: F841
        assert set(cu_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        cu_x: int
        cu_y: float
        cu_x = cu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            cu_y = cu_doc["c"]  # noqa: F841

        # untyped, cursors type inference on find
        c_tb_untyped_cursor = c_tb_untyped.find({}, projection=FIND_PROJECTION)
        cucur_doc = c_tb_untyped_cursor.__next__()
        assert cucur_doc is not None
        cucur_b: int
        cucur_b = cucur_doc["p_bigint"]  # noqa: F841
        assert set(cucur_doc.keys()) == {"p_bigint"}
        # untyped, these are all ok:
        cucur_x: str
        cucur_y: float
        cucur_x = cucur_doc["p_bigint"]  # noqa: F841
        with pytest.raises(KeyError):
            cucur_y = cucur_doc["c"]  # noqa: F841

        # untyped, check using alter table's return value
        altered_c_tb_untyped = c_tb_untyped.alter(ALTER_ADD_COLUMN)
        altered_cu_doc = altered_c_tb_untyped.find_one(FIND_FILTER)
        assert altered_cu_doc is not None
        alt_cu_a: str
        alt_cu_b: int
        alt_cu_a = altered_cu_doc["added"]  # noqa: F841
        alt_cu_b = altered_cu_doc["added"]  # noqa: F841
        altered_c_tb_untyped.alter(ALTER_DROP_COLUMN)

        # Typed
        c_tb_typed: Table[MyTestDoc] = sync_database.create_table(
            TYPING_TABLE_NAME,
            definition=TYPING_TABLE_DEFINITION,
            if_not_exists=True,
            row_type=MyTestDoc,
        )
        c_tb_typed.insert_one(TYPED_ROW)
        ct_doc = c_tb_typed.find_one(FIND_FILTER)
        assert ct_doc is not None
        ct_a: str
        ct_b: int
        ct_a = ct_doc["p_ascii"]  # noqa: F841
        ct_b = ct_doc["p_bigint"]  # noqa: F841
        assert set(ct_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ct_x: int
        ct_y: float
        ct_x = ct_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ct_y = ct_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, cursors type inference on find
        c_tb_typed_cursor = c_tb_typed.find(
            {}, projection=FIND_PROJECTION, row_type=MyTestMiniDoc
        )
        ctcur_doc = c_tb_typed_cursor.__next__()
        assert ctcur_doc is not None
        ctcur_b: int
        ctcur_b = ctcur_doc["p_bigint"]  # noqa: F841
        assert set(ctcur_doc.keys()) == {"p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ctcur_x: str
        ctcur_y: float
        ctcur_x = ctcur_doc["p_bigint"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ctcur_y = ctcur_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, check using alter table's return value
        altered_c_tb_typed = c_tb_typed.alter(
            ALTER_ADD_COLUMN, row_type=AlteredMyTestDoc
        )
        altered_ct_doc = altered_c_tb_typed.find_one(FIND_FILTER)
        assert altered_ct_doc is not None
        alt_ct_a: str
        alt_ct_b: int
        alt_ct_a = altered_ct_doc["added"]  # type: ignore[assignment] # noqa: F841
        alt_ct_b = altered_ct_doc["added"]  # noqa: F841
        altered_c_tb_typed.alter(ALTER_DROP_COLUMN)

    @pytest.mark.describe("test of typing get_table, sync")
    def test_get_table_typing_sync(
        self,
        sync_database: Database,
        sync_typing_test_table: DefaultTable,
    ) -> None:
        """Test of getting typed tables with generics (and not), sync."""

        # Untyped baseline
        g_tb_untyped = sync_database.get_table(TYPING_TABLE_NAME)
        g_tb_untyped.insert_one(ROW)
        gu_doc = g_tb_untyped.find_one(FIND_FILTER)
        assert gu_doc is not None
        gu_a: str
        gu_b: int
        gu_a = gu_doc["p_ascii"]  # noqa: F841
        gu_b = gu_doc["p_bigint"]  # noqa: F841
        assert set(gu_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        gu_x: int
        gu_y: float
        gu_x = gu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            gu_y = gu_doc["c"]  # noqa: F841

        # Typed
        g_tb_typed: Table[MyTestDoc] = sync_database.get_table(
            TYPING_TABLE_NAME, row_type=MyTestDoc
        )
        g_tb_typed.insert_one(TYPED_ROW)
        gt_doc = g_tb_typed.find_one(FIND_FILTER)
        assert gt_doc is not None
        gt_a: str
        gt_b: int
        gt_a = gt_doc["p_ascii"]  # noqa: F841
        gt_b = gt_doc["p_bigint"]  # noqa: F841
        assert set(gt_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        gt_x: int
        gt_y: float
        gt_x = gt_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_y = gt_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

    @pytest.mark.describe("test of typing create_table and alter_table, async")
    async def test_create_alter_table_typing_async(
        self,
        async_database: AsyncDatabase,
        async_typing_test_table: DefaultAsyncTable,
    ) -> None:
        """Test of creating typed tables with generics (and not), async."""

        # Untyped baseline
        ac_tb_untyped = await async_database.create_table(
            TYPING_TABLE_NAME,
            definition=TYPING_TABLE_DEFINITION,
            if_not_exists=True,
        )
        await ac_tb_untyped.insert_one(ROW)
        cu_doc = await ac_tb_untyped.find_one(FIND_FILTER)
        assert cu_doc is not None
        cu_a: str
        cu_b: int
        cu_a = cu_doc["p_ascii"]  # noqa: F841
        cu_b = cu_doc["p_bigint"]  # noqa: F841
        assert set(cu_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        cu_x: int
        cu_y: float
        cu_x = cu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            cu_y = cu_doc["c"]  # noqa: F841

        # untyped, cursors type inference on find
        c_tb_untyped_acursor = ac_tb_untyped.find({}, projection=FIND_PROJECTION)
        cucur_doc = await c_tb_untyped_acursor.__anext__()
        assert cucur_doc is not None
        cucur_b: int
        cucur_b = cucur_doc["p_bigint"]  # noqa: F841
        assert set(cucur_doc.keys()) == {"p_bigint"}
        # untyped, these are all ok:
        cucur_x: str
        cucur_y: float
        cucur_x = cucur_doc["p_bigint"]  # noqa: F841
        with pytest.raises(KeyError):
            cucur_y = cucur_doc["c"]  # noqa: F841

        # untyped, check using alter table's return value
        altered_ac_tb_untyped = await ac_tb_untyped.alter(ALTER_ADD_COLUMN)
        altered_cu_doc = await altered_ac_tb_untyped.find_one(FIND_FILTER)
        assert altered_cu_doc is not None
        alt_cu_a: str
        alt_cu_b: int
        alt_cu_a = altered_cu_doc["added"]  # noqa: F841
        alt_cu_b = altered_cu_doc["added"]  # noqa: F841
        await altered_ac_tb_untyped.alter(ALTER_DROP_COLUMN)

        # Typed
        ac_tb_typed: AsyncTable[MyTestDoc] = await async_database.create_table(
            TYPING_TABLE_NAME,
            definition=TYPING_TABLE_DEFINITION,
            if_not_exists=True,
            row_type=MyTestDoc,
        )
        await ac_tb_typed.insert_one(TYPED_ROW)
        ct_doc = await ac_tb_typed.find_one(FIND_FILTER)
        assert ct_doc is not None
        ct_a: str
        ct_b: int
        ct_a = ct_doc["p_ascii"]  # noqa: F841
        ct_b = ct_doc["p_bigint"]  # noqa: F841
        assert set(ct_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ct_x: int
        ct_y: float
        ct_x = ct_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ct_y = ct_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, cursors type inference on find
        c_tb_typed_acursor = ac_tb_typed.find(
            {}, projection=FIND_PROJECTION, row_type=MyTestMiniDoc
        )
        ctcur_doc = await c_tb_typed_acursor.__anext__()
        assert ctcur_doc is not None
        ctcur_b: int
        ctcur_b = ctcur_doc["p_bigint"]  # noqa: F841
        assert set(ctcur_doc.keys()) == {"p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ctcur_x: str
        ctcur_y: float
        ctcur_x = ctcur_doc["p_bigint"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ctcur_y = ctcur_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, check using alter table's return value
        altered_ac_tb_typed = await ac_tb_typed.alter(
            ALTER_ADD_COLUMN, row_type=AlteredMyTestDoc
        )
        altered_ct_doc = await altered_ac_tb_typed.find_one(FIND_FILTER)
        assert altered_ct_doc is not None
        alt_ct_a: str
        alt_ct_b: int
        alt_ct_a = altered_ct_doc["added"]  # type: ignore[assignment] # noqa: F841
        alt_ct_b = altered_ct_doc["added"]  # noqa: F841
        await altered_ac_tb_typed.alter(ALTER_DROP_COLUMN)

    @pytest.mark.describe("test of typing get_table, async")
    async def test_get_table_typing_async(
        self,
        async_database: AsyncDatabase,
        async_typing_test_table: DefaultAsyncTable,
    ) -> None:
        """Test of getting typed tables with generics (and not), async."""

        # Untyped baseline
        ag_tb_untyped = async_database.get_table(TYPING_TABLE_NAME)
        await ag_tb_untyped.insert_one(ROW)
        gu_doc = await ag_tb_untyped.find_one(FIND_FILTER)
        assert gu_doc is not None
        gu_a: str
        gu_b: int
        gu_a = gu_doc["p_ascii"]  # noqa: F841
        gu_b = gu_doc["p_bigint"]  # noqa: F841
        assert set(gu_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        gu_x: int
        gu_y: float
        gu_x = gu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            gu_y = gu_doc["c"]  # noqa: F841

        # Typed
        ag_tb_typed: AsyncTable[MyTestDoc] = async_database.get_table(
            TYPING_TABLE_NAME, row_type=MyTestDoc
        )
        await ag_tb_typed.insert_one(TYPED_ROW)
        gt_doc = await ag_tb_typed.find_one(FIND_FILTER)
        assert gt_doc is not None
        gt_a: str
        gt_b: int
        gt_a = gt_doc["p_ascii"]  # noqa: F841
        gt_b = gt_doc["p_bigint"]  # noqa: F841
        assert set(gt_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        gt_x: int
        gt_y: float
        gt_x = gt_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_y = gt_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from decimal import Decimal
from typing import Any

from astrapy.utils.api_options import SerdesOptions, defaultSerdesOptions

S_OPTS_NO_DECS = defaultSerdesOptions.with_override(
    SerdesOptions(use_decimals_in_collections=False),
)
S_OPTS_OK_DECS = defaultSerdesOptions.with_override(
    SerdesOptions(use_decimals_in_collections=True),
)
_BASELINE_SCALAR_CASES = {
    "_id": "baseline",
    "f": 1.23,
    "i": 123,
    "t": "T",
}
_W_DECIMALS_SCALAR_CASES = {
    "_id": "decimals",
    "f": 1.23,
    "i": 123,
    "t": "T",
    "de": Decimal("1.23"),
    "dfA": Decimal(1.23),
    "dfB": Decimal("1.229999999999999982236431605997495353221893310546875"),
}
BASELINE_OBJ = {
    **_BASELINE_SCALAR_CASES,
    "subdict": _BASELINE_SCALAR_CASES,
    "sublist": list(_BASELINE_SCALAR_CASES.values()),
}
WDECS_OBJ = {
    **_W_DECIMALS_SCALAR_CASES,
    "subdict": _W_DECIMALS_SCALAR_CASES,
    "sublist": list(_W_DECIMALS_SCALAR_CASES.values()),
}


def is_decimal_super(more_decs: dict[str, Any], less_decs: dict[str, Any]) -> bool:
    """
    Return True if the first item is "the same values, possibly made Decimal
    where the corresponding second item can be another number (int/float)".
    """
    if isinstance(more_decs, list):
        if not isinstance(less_decs, list):
            return False
        if len(more_decs) != len(less_decs):
            return False
        return all(
            [
                is_decimal_super(v_more, v_less)
                for v_more, v_less in zip(more_decs, less_decs)
            ]
        )
    elif isinstance(more_decs, dict):
        if not isinstance(less_decs, dict):
            return False
        if more_decs.keys() != less_decs.keys():
            return False
        return all(
            [is_decimal_super(v_more, less_decs[k]) for k, v_more in more_decs.items()]
        )
    else:
        # other scalars
        if isinstance(more_decs, Decimal):
            if isinstance(less_decs, Decimal):
                return more_decs == less_decs
            else:
                return float(more_decs) == float(less_decs)
        else:
            if isinstance(less_decs, Decimal):
                return False
            return more_decs == less_decs



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.data.utils.extended_json_converters import (
    convert_ejson_binary_object_to_bytes,
    convert_to_ejson_bytes,
)
from astrapy.data_types import DataAPIVector
from astrapy.data_types.data_api_vector import bytes_to_floats, floats_to_bytes

COMPARE_EPSILON = 0.00001

SAMPLE_FLOAT_LISTS: list[list[float]] = [
    [10, 100, 1000, 0.1],
    [0.0043, 0.123, 1.332],
    [],
    [103.1, 104.5, 105.6],
    [(-1 if i % 2 == 0 else +1) * i / 5000 for i in range(4096)],
]


def _nearly_equal_lists(list1: list[float], list2: list[float]) -> bool:
    if len(list1) != len(list2):
        return False
    if len(list1) == 0:
        return True
    return max(abs(x - y) for x, y in zip(list1, list2)) < COMPARE_EPSILON


def _nearly_equal_vectors(vec1: DataAPIVector, vec2: DataAPIVector) -> bool:
    return _nearly_equal_lists(vec1.data, vec2.data)


class TestDataAPIVector:
    @pytest.mark.describe("test of float-binary conversions")
    def test_dataapivector_byteconversions(self) -> None:
        for test_list0 in SAMPLE_FLOAT_LISTS:
            test_bytes0 = floats_to_bytes(test_list0)
            test_list1 = bytes_to_floats(test_bytes0)
            assert _nearly_equal_lists(test_list0, test_list1)

    @pytest.mark.describe("test of float-string conversions")
    def test_dataapivector_stringconversions(self) -> None:
        # known expectation
        obj_1 = {"$binary": "PczMzT5MzM0+mZma"}
        vec = DataAPIVector.from_bytes(convert_ejson_binary_object_to_bytes(obj_1))
        vec_exp = DataAPIVector([0.1, 0.2, 0.3])
        assert _nearly_equal_vectors(vec, vec_exp)
        # some full-round conversions
        for test_list0 in SAMPLE_FLOAT_LISTS:
            test_vec0 = DataAPIVector(test_list0)
            test_ejson = convert_to_ejson_bytes(test_vec0.to_bytes())
            test_vec1 = DataAPIVector.from_bytes(
                convert_ejson_binary_object_to_bytes(test_ejson)
            )
            assert _nearly_equal_vectors(test_vec0, test_vec1)

    @pytest.mark.describe("test of DataAPIVector lifecycle")
    def test_dataapivector_lifecycle(self) -> None:
        v0 = DataAPIVector([])
        v1 = DataAPIVector([1.1, 2.2, 3.3])
        for i in v0:
            pass
        for i in v1:
            pass
        assert list(v0) == []
        assert list(v1) == [1.1, 2.2, 3.3]
        assert v0 != v1
        assert v1 == DataAPIVector([1.1, 2.2, 3.3])

        # list-likeness
        assert v1[1:2] == DataAPIVector([2.2])
        assert len(v1) == 3



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import sys
from collections.abc import Iterable
from typing import Any

from astrapy.info import EmbeddingProviderParameter, VectorServiceOptions
from astrapy.settings.defaults import (
    EMBEDDING_HEADER_API_KEY,
    EMBEDDING_HEADER_AWS_ACCESS_ID,
    EMBEDDING_HEADER_AWS_SECRET_ID,
)

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from live_provider_info import live_provider_info

alphanum = set("qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890")

TEST_EXTENDED_VECTORIZE = bool(os.environ.get("TEST_EXTENDED_VECTORIZE"))

DEFAULT_TEST_ASSETS = {
    "samples": [
        (
            "car",
            "My car is fast and red, it's awesome.",
        ),
        (
            "landscape",
            "The landscape shows spiky, rocky mountaintops and a river nearby...",
        ),
        (
            "motorbike",
            "Your motorbike is certainly the best I have seen!",
        ),
    ],
    "probe": {
        "text": "Driving a red car is a joy.",
        "expected": ["car", "motorbike"],
    },
}

CODE_TEST_ASSETS = {
    "samples": [
        (
            "py_loop",
            "import os; for i in range(10): print(i + os.getenv('env'))",
        ),
        (
            "java",
            (
                "if (sortClause == null || sortClause.sortExpressions()."
                "isEmpty()){return Uni.createFrom().item(true);}"
            ),
        ),
        (
            "py_while",
            (
                "import json; val=json.load(open('limit.json')); "
                "v=0; while v<val: pprint(v); v = v + 1"
            ),
        ),
    ],
    "probe": {
        "text": "for(q in [10, 11, 12]): print(f'q={q}')",
        "expected": ["py_loop", "py_while"],
    },
}

TEST_ASSETS_MAP = {
    ("jinaAI", "jina-embeddings-v2-base-code"): CODE_TEST_ASSETS,
    ("voyageAI", "voyage-code-2"): CODE_TEST_ASSETS,
}

USE_INSERT_ONE_MAP: dict[tuple[str, str], bool] = {
    # ("upstageAI", "solar-1-mini-embedding"): True,
}

SECRET_NAME_ROOT_MAP = {
    "azureOpenAI": "AZURE_OPENAI",
    "bedrock": "BEDROCK",
    "cohere": "COHERE",
    "huggingface": "HUGGINGFACE",
    "huggingfaceDedicated": "HUGGINGFACEDED",
    "jinaAI": "JINAAI",
    "mistral": "MISTRAL",
    "nvidia": "NVIDIA",
    "openai": "OPENAI",
    "upstageAI": "UPSTAGE",
    "voyageAI": "VOYAGEAI",
}

# Any "model tag" listed here is skipped no matter what.
# (useful for models already from the findEmbProvs but not stable/working yet).
EXCLUDED_MODEL_QUADRUPLES = {
    # Cohere provider, while returned by FindEmbProv's, is not in scope for Vectorize and is skipped:
    ("cohere", "embed-english-v2.0", "HEADER", "f"),
    ("cohere", "embed-english-v3.0", "HEADER", "f"),
    ("cohere", "embed-multilingual-v3.0", "HEADER", "f"),
    # Pending a testable deployment, HuggingFace Dedicated testing is momentarily suspended:
    ("huggingfaceDedicated", "endpoint-defined-model", "HEADER", "f"),
    ("huggingfaceDedicated", "endpoint-defined-model", "SHARED_SECRET", "f"),
}

# this is a way to suppress/limit certain combinations of
# "full param" testing from the start. If even one of the optional params
# in a test model is PARAM_SKIP_MARKER, the combination is not emitted at all.
PARAM_SKIP_MARKER = "__SKIP_ME__"

PARAMETER_VALUE_MAP: dict[tuple[str, str, str], Any]
FORCE_DIMENSION_MAP: dict[tuple[str, str], Any]
if TEST_EXTENDED_VECTORIZE:
    PARAMETER_VALUE_MAP = {
        ("azureOpenAI", "text-embedding-3-large", "deploymentId"): os.environ[
            "AZURE_OPENAI_DEPLOY_ID_EMB3LARGE"
        ],
        ("azureOpenAI", "text-embedding-3-large", "resourceName"): os.environ[
            "AZURE_OPENAI_RESNAME_EMB3LARGE"
        ],
        ("azureOpenAI", "text-embedding-3-small", "deploymentId"): os.environ[
            "AZURE_OPENAI_DEPLOY_ID_EMB3SMALL"
        ],
        ("azureOpenAI", "text-embedding-3-small", "resourceName"): os.environ[
            "AZURE_OPENAI_RESNAME_EMB3SMALL"
        ],
        ("azureOpenAI", "text-embedding-ada-002", "deploymentId"): os.environ[
            "AZURE_OPENAI_DEPLOY_ID_ADA2"
        ],
        ("azureOpenAI", "text-embedding-ada-002", "resourceName"): os.environ[
            "AZURE_OPENAI_RESNAME_ADA2"
        ],
        ("jinaAI", "jina-embeddings-v3", "late_chunking"): True,
        ("jinaAI", "jina-embeddings-v3", "task"): "text-matching",
        ("voyageAI", "voyage-2", "autoTruncate"): True,
        ("voyageAI", "voyage-code-2", "autoTruncate"): True,
        ("voyageAI", "voyage-finance-2", "autoTruncate"): True,
        ("voyageAI", "voyage-large-2", "autoTruncate"): True,
        ("voyageAI", "voyage-large-2-instruct", "autoTruncate"): True,
        ("voyageAI", "voyage-law-2", "autoTruncate"): True,
        ("voyageAI", "voyage-multilingual-2", "autoTruncate"): True,
        #
        ("huggingfaceDedicated", "endpoint-defined-model", "endpointName"): os.environ[
            "HUGGINGFACEDED_ENDPOINTNAME"
        ],
        ("huggingfaceDedicated", "endpoint-defined-model", "regionName"): os.environ[
            "HUGGINGFACEDED_REGIONNAME"
        ],
        ("huggingfaceDedicated", "endpoint-defined-model", "cloudName"): os.environ[
            "HUGGINGFACEDED_CLOUDNAME"
        ],
        #
        ("openai", "text-embedding-3-large", "organizationId"): os.environ[
            "OPENAI_ORGANIZATION_ID"
        ],
        ("openai", "text-embedding-3-large", "projectId"): os.environ[
            "OPENAI_PROJECT_ID"
        ],
        ("openai", "text-embedding-3-small", "organizationId"): os.environ[
            "OPENAI_ORGANIZATION_ID"
        ],
        ("openai", "text-embedding-3-small", "projectId"): os.environ[
            "OPENAI_PROJECT_ID"
        ],
        ("openai", "text-embedding-ada-002", "organizationId"): os.environ[
            "OPENAI_ORGANIZATION_ID"
        ],
        ("openai", "text-embedding-ada-002", "projectId"): os.environ[
            "OPENAI_PROJECT_ID"
        ],
        #
        ("bedrock", "amazon.titan-embed-text-v1", "region"): os.environ[
            "BEDROCK_REGION"
        ],
        ("bedrock", "amazon.titan-embed-text-v2:0", "region"): os.environ[
            "BEDROCK_REGION"
        ],
    }

    # this is ad-hoc for HF dedicated. Models here, though "optional" dimension,
    # do not undergo the f/0 optional dimension because of that, rather have
    # a forced fixed, provided dimension.
    FORCE_DIMENSION_MAP = {
        ("huggingfaceDedicated", "endpoint-defined-model"): int(
            os.environ["HUGGINGFACEDED_DIMENSION"]
        ),
    }
else:
    PARAMETER_VALUE_MAP = {}
    FORCE_DIMENSION_MAP = {}


def live_test_models() -> Iterable[dict[str, Any]]:
    def _from_validation(pspec: EmbeddingProviderParameter) -> int:
        assert pspec.parameter_type.upper() == "NUMBER"
        if "numericRange" in pspec.validation:
            m0: int = pspec.validation["numericRange"][0]
            m1: int = pspec.validation["numericRange"][1]
            return (m0 + m1) // 2
        elif "options" in pspec.validation:
            options: list[int] = pspec.validation["options"]
            if len(options) > 1:
                return options[1]
            else:
                return options[0]
        else:
            raise ValueError("unsupported pspec")

    def _collapse(longt: str) -> str:
        if len(longt) <= 40:
            return longt
        else:
            return f"{longt[:30]}_{longt[-5:]}"

    if TEST_EXTENDED_VECTORIZE:
        # generate the full list of models based on the live provider endpoint
        provider_info = live_provider_info()
        for provider_name, provider_desc in sorted(
            provider_info.embedding_providers.items()
        ):
            for model in provider_desc.models:
                for auth_type_name, auth_type_desc in sorted(
                    provider_desc.supported_authentication.items()
                ):
                    if auth_type_desc.enabled:
                        # test assumptions on auth type
                        if auth_type_name == "NONE":
                            assert auth_type_desc.tokens == []
                        elif auth_type_name == "HEADER":
                            header_names_lower = tuple(
                                sorted(
                                    t.accepted.lower() for t in auth_type_desc.tokens
                                )
                            )
                            assert header_names_lower in {
                                (EMBEDDING_HEADER_API_KEY.lower(),),
                                (
                                    EMBEDDING_HEADER_AWS_ACCESS_ID.lower(),
                                    EMBEDDING_HEADER_AWS_SECRET_ID.lower(),
                                ),
                            }
                        elif auth_type_name == "SHARED_SECRET":
                            authkey_names = tuple(
                                sorted(t.accepted for t in auth_type_desc.tokens)
                            )
                            assert authkey_names in {
                                ("providerKey",),
                                ("accessId", "secretKey"),
                            }
                        else:
                            raise ValueError("Unknown auth type")

                        # params
                        collated_params = provider_desc.parameters + model.parameters
                        all_nond_params = [
                            param
                            for param in collated_params
                            if param.name != "vectorDimension"
                        ]
                        required_nond_params = {
                            param.name for param in all_nond_params if param.required
                        }
                        optional_nond_params = {
                            param.name
                            for param in all_nond_params
                            if not param.required
                        }
                        #
                        d_params = [
                            param
                            for param in collated_params
                            if param.name == "vectorDimension"
                        ]
                        if d_params:
                            d_param = d_params[0]
                            if (provider_name, model.name) in FORCE_DIMENSION_MAP:
                                optional_dimension = False
                                dimension = FORCE_DIMENSION_MAP[
                                    (provider_name, model.name)
                                ]
                            elif d_param.default_value is not None:
                                optional_dimension = True
                                assert model.vector_dimension is None
                                dimension = _from_validation(d_param)
                            else:
                                optional_dimension = False
                                assert model.vector_dimension is None
                                dimension = _from_validation(d_param)
                        else:
                            optional_dimension = False
                            assert model.vector_dimension is not None
                            assert model.vector_dimension > 0
                            dimension = model.vector_dimension

                        model_parameters = {
                            param_name: PARAMETER_VALUE_MAP[
                                (provider_name, model.name, param_name)
                            ]
                            for param_name in required_nond_params
                        }
                        optional_model_parameters = {
                            param_name: PARAMETER_VALUE_MAP[
                                (provider_name, model.name, param_name)
                            ]
                            for param_name in optional_nond_params
                        }

                        minimal_model_quadruple = (
                            provider_name,
                            model.name,
                            auth_type_name,
                            "0",
                        )
                        if minimal_model_quadruple not in EXCLUDED_MODEL_QUADRUPLES:
                            if optional_dimension or optional_nond_params != set():
                                # we issue a minimal-params version
                                model_tag_0 = (
                                    f"{provider_name}/{model.name}/{auth_type_name}/0"
                                )
                                this_minimal_model = {
                                    "model_tag": model_tag_0,
                                    "simple_tag": _collapse(
                                        "".join(c for c in model_tag_0 if c in alphanum)
                                    ),
                                    "auth_type_name": auth_type_name,
                                    "auth_type_tokens": auth_type_desc.tokens,
                                    "secret_tag": SECRET_NAME_ROOT_MAP[provider_name],
                                    "test_assets": TEST_ASSETS_MAP.get(
                                        (provider_name, model.name), DEFAULT_TEST_ASSETS
                                    ),
                                    "use_insert_one": USE_INSERT_ONE_MAP.get(
                                        (provider_name, model.name), False
                                    ),
                                    "service_options": VectorServiceOptions(
                                        provider=provider_name,
                                        model_name=model.name,
                                        parameters=model_parameters,
                                    ),
                                }
                                yield this_minimal_model

                        # and in any case we issue a 'full-spec' one ...
                        # ... unless explicitly marked as skipped
                        if all(
                            v != PARAM_SKIP_MARKER
                            for v in optional_model_parameters.values()
                        ):
                            root_model = {
                                "auth_type_name": auth_type_name,
                                "auth_type_tokens": auth_type_desc.tokens,
                                "dimension": dimension,
                                "secret_tag": SECRET_NAME_ROOT_MAP[provider_name],
                                "test_assets": TEST_ASSETS_MAP.get(
                                    (provider_name, model.name), DEFAULT_TEST_ASSETS
                                ),
                                "use_insert_one": USE_INSERT_ONE_MAP.get(
                                    (provider_name, model.name), False
                                ),
                            }

                            model_tag_f = (
                                f"{provider_name}/{model.name}/{auth_type_name}/f"
                            )
                            full_model_quadruple = (
                                provider_name,
                                model.name,
                                auth_type_name,
                                "f",
                            )
                            if full_model_quadruple not in EXCLUDED_MODEL_QUADRUPLES:
                                this_model = {
                                    "model_tag": model_tag_f,
                                    "simple_tag": _collapse(
                                        "".join(c for c in model_tag_f if c in alphanum)
                                    ),
                                    "service_options": VectorServiceOptions(
                                        provider=provider_name,
                                        model_name=model.name,
                                        parameters={
                                            **model_parameters,
                                            **optional_model_parameters,
                                        },
                                    ),
                                    **root_model,
                                }
                                yield this_model



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy import AsyncDatabase
from astrapy.constants import ModelStatus
from astrapy.info import FindRerankingProvidersResult, RerankingProvider


def _count_models(frp_result: FindRerankingProvidersResult) -> int:
    return len(
        [
            model
            for prov_v in frp_result.reranking_providers.values()
            for model in prov_v.models
        ]
    )


class TestRerankingOpsAsync:
    @pytest.mark.describe("test of find_reranking_providers, async")
    async def test_findrerankingproviders_async(
        self,
        async_database: AsyncDatabase,
    ) -> None:
        database_admin = async_database.get_database_admin()
        rp_result = await database_admin.async_find_reranking_providers()

        assert isinstance(rp_result, FindRerankingProvidersResult)

        assert all(
            isinstance(rer_prov, RerankingProvider)
            for rer_prov in rp_result.reranking_providers.values()
        )

        # 'raw_info' not compared, for resiliency against newly-introduced fields
        assert (
            FindRerankingProvidersResult._from_dict(
                rp_result.as_dict()
            ).reranking_providers
            == rp_result.reranking_providers
        )

    @pytest.mark.describe("test of find_reranking_providers filtering, async")
    async def test_filtered_findrerankingproviders_async(
        self,
        async_database: AsyncDatabase,
    ) -> None:
        database_admin = async_database.get_database_admin()
        default_count = _count_models(
            await database_admin.async_find_reranking_providers()
        )

        all_count = _count_models(
            await database_admin.async_find_reranking_providers(filter_model_status="")
        )
        all_count_e = _count_models(
            await database_admin.async_find_reranking_providers(
                filter_model_status=ModelStatus.ALL,
            )
        )
        assert all_count_e == all_count

        sup_count = _count_models(
            await database_admin.async_find_reranking_providers(
                filter_model_status="SUPPORTED"
            )
        )
        dep_count = _count_models(
            await database_admin.async_find_reranking_providers(
                filter_model_status="DEPRECATED"
            )
        )
        eol_count = _count_models(
            await database_admin.async_find_reranking_providers(
                filter_model_status="END_OF_LIFE"
            )
        )
        sup_count_e = _count_models(
            await database_admin.async_find_reranking_providers(
                filter_model_status=ModelStatus.SUPPORTED,
            )
        )
        dep_count_e = _count_models(
            await database_admin.async_find_reranking_providers(
                filter_model_status=ModelStatus.DEPRECATED,
            )
        )
        eol_count_e = _count_models(
            await database_admin.async_find_reranking_providers(
                filter_model_status=ModelStatus.END_OF_LIFE,
            )
        )
        assert sup_count_e == sup_count
        assert dep_count_e == dep_count
        assert eol_count_e == eol_count

        assert sup_count + dep_count + eol_count == all_count
        assert sup_count == default_count



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import argparse
import logging
import os
import sys
from code import interact
from typing import Any

from astrapy import (
    AstraDBAdmin,
    AstraDBDatabaseAdmin,
    AsyncCollection,
    AsyncDatabase,
    AsyncTable,
    Collection,
    DataAPIClient,
    DataAPIDatabaseAdmin,
    Database,
    Table,
)
from astrapy.admin import (
    ParsedAPIEndpoint,
    parse_api_endpoint,
)
from astrapy.api_options import (
    APIOptions,
    DataAPIURLOptions,
    DevOpsAPIURLOptions,
    SerdesOptions,
    TimeoutOptions,
)
from astrapy.authentication import (
    AWSEmbeddingHeadersProvider,
    EmbeddingAPIKeyHeaderProvider,
    StaticTokenProvider,
    TokenProvider,
    UsernamePasswordTokenProvider,
)
from astrapy.constants import (
    DefaultIdType,
    Environment,
    MapEncodingMode,
    ReturnDocument,
    SortMode,
    VectorMetric,
)
from astrapy.cursors import (
    AbstractCursor,
    AsyncCollectionFindAndRerankCursor,
    AsyncCollectionFindCursor,
    AsyncTableFindCursor,
    CollectionFindAndRerankCursor,
    CollectionFindCursor,
    CursorState,
    RerankedResult,
    TableFindCursor,
)
from astrapy.data_types import (
    DataAPIDate,
    DataAPIDictUDT,
    DataAPIDuration,
    DataAPIMap,
    DataAPISet,
    DataAPITime,
    DataAPITimestamp,
    DataAPIVector,
)
from astrapy.event_observers import (
    ObservableError,
    ObservableEvent,
    ObservableEventType,
    ObservableRequest,
    ObservableResponse,
    ObservableWarning,
    Observer,
    event_collector,
)
from astrapy.ids import (
    UUID,
    ObjectId,
    uuid1,
    uuid3,
    uuid4,
    uuid5,
    uuid6,
    uuid7,
    uuid8,
)
from astrapy.info import (
    AlterTableAddColumns,
    AlterTableAddVectorize,
    AlterTableDropColumns,
    AlterTableDropVectorize,
    AlterTypeAddFields,
    AlterTypeOperation,
    AlterTypeRenameFields,
    AstraDBAdminDatabaseInfo,
    AstraDBDatabaseInfo,
    CollectionDefaultIDOptions,
    CollectionDefinition,
    CollectionDescriptor,
    CollectionInfo,
    CollectionLexicalOptions,
    CollectionRerankOptions,
    CollectionVectorOptions,
    ColumnType,
    CreateTableDefinition,
    CreateTypeDefinition,
    EmbeddingProvider,
    EmbeddingProviderAuthentication,
    EmbeddingProviderModel,
    EmbeddingProviderParameter,
    EmbeddingProviderToken,
    FindEmbeddingProvidersResult,
    FindRerankingProvidersResult,
    ListTableDefinition,
    ListTableDescriptor,
    ListTypeDescriptor,
    RerankingProvider,
    RerankingProviderAuthentication,
    RerankingProviderModel,
    RerankingProviderParameter,
    RerankingProviderToken,
    RerankServiceOptions,
    TableAPIIndexSupportDescriptor,
    TableAPISupportDescriptor,
    TableBaseIndexDefinition,
    TableIndexDefinition,
    TableIndexDescriptor,
    TableIndexOptions,
    TableInfo,
    TableKeyValuedColumnType,
    TableKeyValuedColumnTypeDescriptor,
    TablePrimaryKeyDescriptor,
    TableScalarColumnTypeDescriptor,
    TableTextIndexDefinition,
    TableTextIndexOptions,
    TableUDTColumnDescriptor,
    TableUnsupportedColumnTypeDescriptor,
    TableUnsupportedIndexDefinition,
    TableValuedColumnType,
    TableValuedColumnTypeDescriptor,
    TableVectorColumnTypeDescriptor,
    TableVectorIndexDefinition,
    TableVectorIndexOptions,
    VectorServiceOptions,
)
from astrapy.utils.api_options import defaultAPIOptions
from astrapy.utils.document_paths import (
    escape_field_names,
    unescape_field_path,
)

ipython_available: bool
try:
    from IPython import embed

    ipython_available = True
except ImportError:
    ipython_available = False

BANNER_TEMPLATE = """*******************************************************************
*
*    █████╗ ███████╗████████╗██████╗  █████╗ ██████╗ ██╗   ██╗
*   ██╔══██╗██╔════╝╚══██╔══╝██╔══██╗██╔══██╗██╔══██╗╚██╗ ██╔╝
*   ███████║███████╗   ██║   ██████╔╝███████║██████╔╝ ╚████╔╝
*   ██╔══██║╚════██║   ██║   ██╔══██╗██╔══██║██╔═══╝   ╚██╔╝
*   ██║  ██║███████║   ██║   ██║  ██║██║  ██║██║        ██║
*   ╚═╝  ╚═╝╚══════╝   ╚═╝   ╚═╝  ╚═╝╚═╝  ╚═╝╚═╝        ╚═╝
*
*   ██████╗ ███████╗██████╗ ██╗
*   ██╔══██╗██╔════╝██╔══██╗██║
*   ██████╔╝█████╗  ██████╔╝██║
*   ██╔══██╗██╔══╝  ██╔═══╝ ██║
*   ██║  ██║███████╗██║     ███████╗
*   ╚═╝  ╚═╝╚══════╝╚═╝     ╚══════╝
*
* astrapy-repl - an interactive astrapy Python shell.
*
*     Variables 'client' and 'database' are ready to use.
*     All astrapy useful symbols are imported already.
*
* Targeting:
*  {endpoint:59.59}
*  keyspace={keyspace}, environment={environment}
*******************************************************************"""

PRIVATE_GLOBALS = {
    "argparse",
    "BANNER_TEMPLATE",
    "embed",
    "interact",
    "ipython_available",
    "main",
    "os",
    "parser",
    "sys",
}

LOGGING_LEVELS = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}

parser = argparse.ArgumentParser(
    description=(
        "Interactive REPL for astrapy. Supplied parameters take precedence "
        "over environment variables."
    ),
)
# DB connection
parser.add_argument(
    "--endpoint",
    "-e",
    type=str,
    help=(
        "Data API endpoint (Astra or other deployment). "
        "Env vars: ASTRA_DB_API_ENDPOINT, LOCAL_DATA_API_ENDPOINT"
    ),
)
parser.add_argument(
    "--token",
    "-t",
    type=str,
    help="Database token (alternative to user/password). Env var: ASTRA_DB_APPLICATION_TOKEN",
)
parser.add_argument(
    "--username",
    "-u",
    type=str,
    help="Username (alternative to token; requires password). Env var: LOCAL_DATA_API_USERNAME",
)
parser.add_argument(
    "--password",
    "-p",
    type=str,
    help="Password (alternative to token; requires username). Env var: LOCAL_DATA_API_PASSWORD",
)
parser.add_argument(
    "--keyspace",
    "-k",
    type=str,
    help="Targeted keyspace. Env vars: ASTRA_DB_KEYSPACE, LOCAL_DATA_API_KEYSPACE",
)
parser.add_argument(
    "--environment",
    choices=Environment.values,
    help="Target environment. Usually auto-detected from endpoint. ",
)
# Logger error level
parser.add_argument(
    "--log-level",
    "-l",
    type=str.upper,
    dest="loglevel",
    choices=[
        "DEBUG",
        "INFO",
        "WARNING",
        "ERROR",
        "CRITICAL",
    ],
    help="Logging level. Use 'DEBUG' for comprehensive inspection of requests and responses.",
)
# REPL type
parser.add_argument(
    "--repl",
    choices=["stdlib", "ipython"],
    help="REPL mode (interactive Python shell) to use",
)


def main() -> None:
    args = parser.parse_args()

    # Manual validation and integration with env. variables.
    # We uniformly apply (1) parameters over env vars, (2) astra+local,
    # and then check for conflicts and re-parse endpoint to find out the target env.

    _endpoint = (
        args.endpoint
        or os.getenv("ASTRA_DB_API_ENDPOINT")
        or os.getenv("LOCAL_DATA_API_ENDPOINT")
    )
    _keyspace = (
        args.keyspace
        or os.getenv("ASTRA_DB_KEYSPACE")
        or os.getenv("LOCAL_DATA_API_KEYSPACE")
        or None
    )
    _environment = args.environment or None
    #
    _token_provider: TokenProvider
    _token = args.token or os.getenv("ASTRA_DB_APPLICATION_TOKEN")
    _username = args.username or os.getenv("LOCAL_DATA_API_USERNAME")
    _password = args.password or os.getenv("LOCAL_DATA_API_PASSWORD")
    if _token and (_username or _password):
        parser.error("Specify either a token or username/password pair, not both.")
    if (_username and not _password) or (_password and not _username):
        parser.error("Username and password must must be provided together.")
    if _username and _password:
        _token_provider = UsernamePasswordTokenProvider(_username, _password)
    else:
        _token_provider = StaticTokenProvider(_token)

    if not _endpoint:
        raise SystemExit(
            "The endpoint is required, either via env. var or command-line argument. "
            "Hint: '-h' for help."
        )

    # instantiate client and database
    if not _environment:
        parsed_endpoint = parse_api_endpoint(_endpoint)
        if parsed_endpoint:
            _environment = parsed_endpoint.environment
        else:
            _environment = Environment.HCD
    # inject the token at client-level for repl users to easily spawn admins and such
    client = DataAPIClient(
        environment=_environment,
        token=_token_provider,
    )
    database = client.get_database(  # noqa: F841
        _endpoint,
        keyspace=_keyspace,
    )

    # pick the target REPL mode
    ipython_repl: bool
    if args.repl == "ipython":
        if ipython_available:
            ipython_repl = True
        else:
            raise SystemExit(
                "'IPython' package not found: cannot start the required REPL. "
                "Hint: '-h' for help."
            )
    elif args.repl == "stdlib":
        ipython_repl = False
    else:
        # unspecified repl choice, pick best default
        ipython_repl = ipython_available

    banner = BANNER_TEMPLATE.format(
        endpoint=_endpoint,
        keyspace=_keyspace or "(default)",
        environment=_environment,
    )

    # spawn the repl
    namespace: dict[str, Any] = {
        **{k: v for k, v in globals().items() if k not in PRIVATE_GLOBALS},
        **{
            "client": client,
            "database": database,
        },
    }
    if args.loglevel:
        logging.basicConfig(level=LOGGING_LEVELS[args.loglevel])
    if ipython_repl:
        embed(
            banner1=banner,
            header=" ",
            user_ns=namespace,
            colors="Neutral",
        )  # type: ignore[no-untyped-call]
    else:
        stdbanner = banner + ("\n" * 3)
        if not sys.flags.interactive:
            interact(
                banner=stdbanner,
                local=namespace,
                exitmsg="Closing astrapy-repl.",
            )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time
from typing import TYPE_CHECKING

from astrapy.data_types import DataAPITimestamp
from astrapy.ids import UUID
from astrapy.info import CreateTypeDefinition

if TYPE_CHECKING:
    from cassandra.cluster import Session


SCHEMA_PROPAGATION_DELAY_SECONDS = 1.0

TABLE_NAME_COUNTER = "test_table_counter"
CREATE_TABLE_COUNTER = (
    f"CREATE TABLE {TABLE_NAME_COUNTER} (a TEXT PRIMARY KEY, col_counter COUNTER);"
)
DROP_TABLE_COUNTER = f"DROP TABLE {TABLE_NAME_COUNTER};"
INSERTS_TABLE_COUNTER = [
    f"UPDATE {TABLE_NAME_COUNTER} SET col_counter=col_counter+137 WHERE a='a';",
    f"UPDATE {TABLE_NAME_COUNTER} SET col_counter=col_counter+1 WHERE a='z';",
]
FILTER_COUNTER = {"a": "a"}
EXPECTED_ROW_COUNTER = {"a": "a", "col_counter": 137}

TABLE_NAME_LOWSUPPORT = "test_table_lowsupport"
CREATE_TABLE_LOWSUPPORT = f"""CREATE TABLE {TABLE_NAME_LOWSUPPORT} (
  a TEXT,
  b TEXT,
  col_timeuuid TIMEUUID,
  col_static_timestamp TIMESTAMP STATIC,
  col_static_list LIST<INT> STATIC,
  col_static_list_exotic LIST<BLOB> STATIC,
  col_static_set SET<INT> STATIC,
  col_static_set_exotic SET<BLOB> STATIC,
  col_static_map MAP<INT, TEXT> STATIC,
  col_static_map_exotic MAP<BLOB,BLOB> STATIC,
  col_unsupported FROZEN<LIST<FROZEN<MAP<FROZEN<SET<FLOAT>>, SMALLINT>>>>,
  PRIMARY KEY ((A),B)
);"""
DROP_TABLE_LOWSUPPORT = f"DROP TABLE {TABLE_NAME_LOWSUPPORT};"
INSERTS_TABLE_LOWSUPPORT = [
    (
        f"INSERT INTO {TABLE_NAME_LOWSUPPORT}"
        "    ("
        "    a,"
        "    b,"
        "    col_timeuuid,"
        "    col_static_timestamp,"
        "    col_static_list,"
        "    col_static_list_exotic,"
        "    col_static_set,"
        "    col_static_set_exotic,"
        "    col_static_map,"
        "    col_static_map_exotic,"
        "    col_unsupported"
        ") VALUES ("
        "    'a',"
        "    'b',"
        "    caba4d50-5078-11f0-b63e-bf76cc9ee780,"
        "    '2022-01-01T12:34:56.000',"
        "    [1, 2, 3],"
        "    [0xff, 0xff],"
        "    {1, 2, 3},"
        "    {0xff, 0xff},"
        "    {1: 'one'},"
        "    {0xff: 0xff},"
        "    [{{0.1, 0.2}: 3}, {{0.3, 0.4}: 6}]"
        ");"
    ),
]
FILTER_LOWSUPPORT = {"a": "a", "b": "b"}
ILLEGAL_PROJECTIONS_LOWSUPPORT = [
    {"col_unsupported": True},
]
PROJECTION_LOWSUPPORT = {"col_unsupported": False}
EXPECTED_ROW_LOWSUPPORT = {
    "a": "a",
    "b": "b",
    "col_timeuuid": UUID("caba4d50-5078-11f0-b63e-bf76cc9ee780"),
    "col_static_list_exotic": [
        {"$binary": "/w=="},
        {"$binary": "/w=="},
    ],
    "col_static_set": [1, 2, 3],
    "col_static_map_exotic": [
        [
            {"$binary": "/w=="},
            {"$binary": "/w=="},
        ]
    ],
    "col_static_timestamp": DataAPITimestamp.from_string("2022-01-01T12:34:56Z"),
    "col_static_map": [[1, "one"]],
    "col_static_list": [1, 2, 3],
    "col_static_set_exotic": [{"$binary": "/w=="}],
}
LOWSUPPORT_TIMEUUID_STRING = "dd5907d0-520c-11f0-a406-bf76cc9ee780"
LOWSUPPORT_TIMEUUID_PK0 = {
    "a": "xA0",
    "b": "xB0",
}
LOWSUPPORT_TIMEUUID_DOC0 = {
    "col_timeuuid": UUID(LOWSUPPORT_TIMEUUID_STRING),
    **LOWSUPPORT_TIMEUUID_PK0,
}
LOWSUPPORT_TIMEUUID_PK1 = {
    "a": "xA1",
    "b": "xB1",
}
LOWSUPPORT_TIMEUUID_DOC1 = {
    "col_timeuuid": LOWSUPPORT_TIMEUUID_STRING,
    **LOWSUPPORT_TIMEUUID_PK1,
}


def _extract_udt_definition(
    session: Session,
    keyspace: str,
    udt_name: str,
    allow_schema_propagation_delay: bool = True,
) -> CreateTypeDefinition | None:
    if allow_schema_propagation_delay:
        time.sleep(SCHEMA_PROPAGATION_DELAY_SECONDS)

    metadata_type = session.cluster.metadata.keyspaces[keyspace].user_types.get(
        udt_name
    )
    if metadata_type is None:
        return None

    fields = dict(zip(metadata_type.field_names, metadata_type.field_types))

    return CreateTypeDefinition(fields=dict(fields))
